import kaa_testutils
from kaa.filetype.css import cssmode


class TestCSSHighlight(kaa_testutils._TestDocBase):
    DEFAULTMODECLASS = cssmode.CSSMode

    def test_propvalue(self):
        doc = self._getdoc('sel{abc:.1em;def:ghi}')
        doc.mode.run_tokenizer(None)
        tokenizer = cssmode.CSSMode.tokenizer

        kaa_testutils.check_style(doc, 0, 21, 
            [tokenizer.tokens.default] * 3 +
            [tokenizer.tokens.ruleset] * 1 +
            [tokenizer.PropTokenizer.tokens.propname] * 4 +
            [tokenizer.PropTokenizer.PropValueTokenizer.tokens.number] * 4 +
            [tokenizer.PropTokenizer.PropValueTokenizer.tokens.terminate_value] * 1 +
            [tokenizer.PropTokenizer.tokens.propname] * 4 +
            [tokenizer.PropTokenizer.PropValueTokenizer.tokens.string] * 3 +
            [tokenizer.PropTokenizer.tokens.terminate_name] * 1
        )

    def test_no_media(self):
        doc = self._getdoc('sel{abc:def;}')
        doc.mode.run_tokenizer(None)
        tokenizer = cssmode.CSSMode.tokenizer

        kaa_testutils.check_style(doc, 0, 13, 
            [tokenizer.tokens.default] * 3 +
            [tokenizer.tokens.ruleset] * 1 +
            [tokenizer.PropTokenizer.tokens.propname] * 4 +
            [tokenizer.PropTokenizer.PropValueTokenizer.tokens.string] * 3 +
            [tokenizer.PropTokenizer.PropValueTokenizer.tokens.terminate_value] * 1 +
            [tokenizer.PropTokenizer.tokens.terminate_name] * 1
        )

    def test_media(self):
        doc = self._getdoc('@media a{b{c:d}}@media e{f{g:h}}')
        doc.mode.run_tokenizer(None)

        tokenizer = cssmode.CSSMode.tokenizer
        doc.mode.run_tokenizer(None)
        kaa_testutils.check_style(doc, 0, 32, 
            [tokenizer.tokens.media] * 9 +
            [tokenizer.MediaCSSTokenizer.tokens.default] * 1 +
            [tokenizer.MediaCSSTokenizer.tokens.ruleset] * 1 +
            [tokenizer.MediaCSSTokenizer.PropTokenizer.tokens.propname] * 2 +
            [tokenizer.MediaCSSTokenizer.PropTokenizer.PropValueTokenizer.tokens.string] * 1 +
            [tokenizer.MediaCSSTokenizer.PropTokenizer.tokens.terminate_name] * 1 +
            [tokenizer.MediaCSSTokenizer.tokens.terminate_media] * 1 +

            [tokenizer.tokens.media] * 9 +
            [tokenizer.MediaCSSTokenizer.tokens.default] * 1 +
            [tokenizer.MediaCSSTokenizer.tokens.ruleset] * 1 +
            [tokenizer.MediaCSSTokenizer.PropTokenizer.tokens.propname] * 2 +
            [tokenizer.MediaCSSTokenizer.PropTokenizer.PropValueTokenizer.tokens.string] * 1 +
            [tokenizer.MediaCSSTokenizer.PropTokenizer.tokens.terminate_name] * 1 +
            [tokenizer.MediaCSSTokenizer.tokens.terminate_media] * 1
            )



from unittest.mock import patch
import kaa_testutils
from kaa.filetype.default import defaultmode, modebase


class TestDefaultMode(kaa_testutils._TestScreenBase):

    def test_getwords(self):
        doc = self._getdoc('0123   \tabc ::::あいうえおカキクケコ[[')

        assert (
            (0, 4, '0123', 'L_WORDCHAR'),
            (4, 8, '   \t', 'Z_WHITESPACE'),
            (8, 11, 'abc', 'L_WORDCHAR'),
            (11, 12, ' ', 'Z_WHITESPACE'),
            (12, 16, '::::', 'P'),
            (16, 21, 'あいうえお', 'L_HIRAGANA'),
            (21, 26, 'カキクケコ', 'L_KATAKANA'),
            (26, 28, '[[', 'P'),) == tuple(doc.mode.split_word(0))

        doc = self._getdoc('\n\n')

        assert (
            (0, 1, '\n', '_LF'),
        ) == tuple(doc.mode.split_word(0))

    def test_getwordat(self):
                           #0123456789012 3 4 56789012
        doc = self._getdoc('abc 0123 abcあいうdef  ')

        assert doc.mode.get_word_at(0) == (0, 3, 'L_WORDCHAR')
        assert doc.mode.get_word_at(3) == (0, 3, 'L_WORDCHAR')
        assert doc.mode.get_word_at(4) == (4, 8, 'L_WORDCHAR')
        assert doc.mode.get_word_at(8) == (4, 8, 'L_WORDCHAR')
        assert doc.mode.get_word_at(9) == (9, 12, 'L_WORDCHAR')
        assert doc.mode.get_word_at(11) == (9, 12, 'L_WORDCHAR')
        assert doc.mode.get_word_at(12) == (12, 15, 'L_HIRAGANA')
        assert doc.mode.get_word_at(15) == (15, 18, 'L_WORDCHAR')
        assert doc.mode.get_word_at(17) == (15, 18, 'L_WORDCHAR')
        assert doc.mode.get_word_at(18) == (15, 18, 'L_WORDCHAR')
        assert doc.mode.get_word_at(19) is None

    def test_search_next(self):
        w = self._getwnd('abcdefgabcdefg/efg/EFG')

        opt = modebase.SearchOption()
        opt.text = 'EFG'
        opt.ignorecase = True
        opt.word = False
        opt.regex = False

        ret = w.document.mode.search_next(0, opt)
        assert ret.span() == (4, 7)

        opt.word = True
        ret = w.document.mode.search_next(0, opt)
        assert ret.span() == (15, 18)

        opt.ignorecase = False
        ret = w.document.mode.search_next(0, opt)
        assert ret.span() == (19, 22)

        opt.text = r'[EFG]+'
        opt.regex = True
        ret = w.document.mode.search_next(0, opt)
        assert ret.span() == (19, 22)

        opt.text = r'[EFG]+'
        opt.regex = False
        ret = w.document.mode.search_next(0, opt)
        assert ret is None

    def test_search_prev(self):
        w = self._getwnd('abcdefgabcdefg/efg/EFG01234567890')

        opt = modebase.SearchOption()
        opt.text = 'EFG'
        opt.ignorecase = True
        opt.word = False
        opt.regex = False

        endpos = w.document.endpos()
        ret = w.document.mode.search_prev(endpos, opt).span()
        assert ret == (19, 22)

        opt.text = r'[EFG]+'
        opt.regex = False
        ret = w.document.mode.search_prev(endpos, opt)
        assert ret is None

    def test_iter_parenthsis(self):
        doc = self._getdoc("({[]})")
        doc.styles.replaceints(0, doc.endpos(), [1, 2, 3, 4, 5, 6])

        ret = [(pos, c, attr) for pos, c, attr in doc.mode.iter_parenthesis(0)]

        assert ret == [(0, '(', 1), (1, '{', 2), (2, '[', 3),
                       (3, ']', 4), (4, '}', 5), (5, ')', 6)]

    def test_iter_rev_parenthsis(self):
        doc = self._getdoc("({[]})")
        doc.styles.replaceints(0, doc.endpos(), [1, 2, 3, 4, 5, 6])

        ret = [(pos, c, attr)
               for pos, c, attr in doc.mode.iter_rev_parenthesis(5)]

        assert ret == [(5, ')', 6), (4, '}', 5), (3, ']', 4),
                       (2, '[', 3), (1, '{', 2), (0, '(', 1)]

    def test_find_pair_parenthesis(self):
        doc = self._getdoc("({[({[]})]})")
        doc.styles.replaceints(
            0, doc.endpos(), [1, 1, 1, 2, 2, 2, 2, 2, 2, 1, 1, 1])

        assert doc.mode.find_match_parenthesis(0) == 11

    def test_auto_indent(self):
        script = '''
a
'''
        wnd = self._getwnd(script)
        with patch.object(wnd.cursor, 'pos', new=2):
            wnd.document.mode.on_auto_indent(wnd)
        assert wnd.document.gettext(0, wnd.document.endpos()) == '\na\n\n'

        script = '''
    a
'''
        wnd = self._getwnd(script)
        with patch.object(wnd.cursor, 'pos', new=6):
            wnd.document.mode.on_auto_indent(wnd)
        assert wnd.document.gettext(0, wnd.document.endpos()
                                    ) == '\n    a\n    \n'

        script = '''
    a
'''
        wnd = self._getwnd(script)
        with patch.object(wnd.cursor, 'pos', new=5):
            wnd.document.mode.on_auto_indent(wnd)
        assert wnd.document.gettext(0, wnd.document.endpos()
                                    ) == '\n\n    a\n'



from contextlib import ExitStack
from unittest.mock import patch

import kaa
import kaa_testutils


class TestSaveAll(kaa_testutils._TestScreenBase):

    def test_saveall(self):
        with ExitStack() as st:
            ec = st.enter_context

            wnd = self._getwnd('')
            ec(patch.object(wnd.document.undo, 'is_dirty', return_value=True))

            frame = kaa_testutils.DmyFrame()
            ec(patch.object(frame, 'get_editors',
               create=True, return_value=[wnd]))

            mainframe = ec(patch.object(wnd, 'mainframe', create=True))
            mainframe.childframes = [frame]

            ec(patch.object(kaa.app, 'storage', create=True))
            ec(patch.object(kaa.app, 'get_activeframe', create=True,
                            return_value=frame))

            ec(patch.object(wnd.document.undo, 'is_dirty', return_value=True))
            ec(patch.object(wnd.document,
               'get_filename', return_value='filename'))

            kaa.app.file_commands.file_saveall(wnd)

            args, kwargs = kaa.app.storage.save_document.call_args
            assert args[0] is wnd.document



import re
from kaa.filetype.default import defaultmode
from kaa.theme import Style
from kaa import encodingdef
from kaa.syntax_highlight import *

# todo: highlighter should be written.

CSSThemes = {
    'basic': [
        Style('media-selector', 'yellow', 'default'),
        Style('css-selector', 'magenta', 'default'),
        Style('css-propname', 'cyan', 'default', bold=True),
        Style('css-propvalue', 'green', 'default', bold=True),
    ],
}


class MediaToken(SingleToken):

    def on_start(self, doc, match):
        pos, terminates = yield from super().on_start(doc, match)
        pos = yield from self.tokenizer.MediaCSSTokenizer.run(doc, pos)

        return pos, False


class RuleSetToken(SingleToken):

    def on_start(self, doc, match):
        pos, terminates = yield from super().on_start(doc, match)
        if match.group(0) == '{':
            pos = yield from self.tokenizer.PropTokenizer.run(doc, pos)

        return pos, False


class PropNameToken(SingleToken):

    def on_start(self, doc, match):
        pos, terminates = yield from super().on_start(doc, match)
        pos = yield from self.tokenizer.PropValueTokenizer.run(doc, pos)

        return pos, self.terminates


def make_prop_tokenizer(root, terminates=None):
    ret = Tokenizer(parent=root, terminates=terminates,
                    tokens=(
                        ('comment1', Span('comment', r'/\*', '\*/', escape='\\')),
                        ('propname', PropNameToken(
                            'css-propname', [r'[^:\s]+:'])),
                        ('terminate_name', SingleToken(
                            'css-selector', [r'}'], terminates=True)),
                    ))

    ret.PropValueTokenizer = Tokenizer(parent=ret, terminates=terminates,
                                       tokens=(
                                           ('close', Terminator(
                                               'css-selector', [r'\}'])),
                                           ('terminate_value', SingleToken(
                                               'css-propvalue', [r';'], terminates=True)),
                                           ('comment1', Span('comment',
                                                             r'/\*', '\*/', escape='\\')),
                                           ('string', SingleToken(
                                               'string', [r'[a-zA-Z_]\w*'])),
                                           ('string1', Span(
                                               'string', '"', '"', escape='\\')),
                                           ('string2', Span(
                                               'string', "'", "'", escape='\\')),
                                           ("color", SingleToken(
                                               'keyword', [r'\#[0-9a-zA-Z]+'])),
                                           ("number",
                                              SingleToken('number',
                                                          [r'[+-]?[0-9]+(\.[0-9]*)*([a-z]*)', r'[+-]?\.[0-9]([a-z]*)'])),
                                       ))

    return ret


def get_css_tokens():
    return [
        ('comment1', Span('comment', r'/\*', '\*/', escape='\\')),
        ('string1', Span('string', '"', '"', escape='\\')),
        ('string2', Span('string', "'", "'", escape='\\')),
        ('ruleset', RuleSetToken('css-selector', [r'\{'])),
    ]


def make_tokenizer(parent, terminates=None):
    ret = Tokenizer(parent=parent, default_style='css-selector',
                    terminates=terminates,
                    tokens=[
                        ('media', MediaToken(
                            'media-selector', [r'@media[^{]*\{?'])),
                        ('decl', SingleToken('directive', [r'@\w*'])),
                    ]
                    + get_css_tokens())

    ret.PropTokenizer = make_prop_tokenizer(ret, terminates=terminates)
    ret.MediaCSSTokenizer = Tokenizer(
        parent=ret, default_style='css-selector',
        tokens=get_css_tokens() +
        [('terminate_media', SingleToken('media-selector', [r'}'], terminates=True))])

    ret.MediaCSSTokenizer.PropTokenizer = make_prop_tokenizer(ret.MediaCSSTokenizer,
                                                              terminates=terminates)
    return ret


class CSSMode(defaultmode.DefaultMode):
    MODENAME = 'CSS'
    tokenizer = make_tokenizer(None, None)

    @classmethod
    def update_fileinfo(cls, fileinfo, document=None):
        if not document:
            try:
                with open(fileinfo.fullpathname, 'rb') as buffer:
                    s = buffer.read(1024)
            except IOError:
                return
        else:
            s = document.gettext(0, 1024).encode('utf-8', errors='ignore')

        m = re.search(
            rb'@charset\s+(' + rb'"([^"]+)"|' + rb"('[^']+'))", s)
        if m:
            enc = str(m.group(2) or m.group(3), 'utf-8', 'replace').strip()
            enc = encodingdef.normalize_encname(enc, fileinfo.encoding)
            fileinfo.encoding = enc

    def init_themes(self):
        super().init_themes()
        self.themes.append(CSSThemes)



from pytest import raises
import curses
import kaa
from kaa.cui import app
from kaa import keyboard
import kaa.options
from kaa.exceptions import KaaError


class TestKey:

    def test_keytrans(self):
        cuiapp = app.CuiApp(kaa.options.build_parser().parse_args([]))
        assert 'A' == cuiapp.translate_key((), 'A')
        assert [curses.KEY_RIGHT] == cuiapp.translate_key((), keyboard.right)
        with raises(KaaError):
            cuiapp.translate_key((), keyboard.insert)

    def test_shift(self):
        cuiapp = app.CuiApp(kaa.options.build_parser().parse_args([]))
        with raises(KaaError):
            cuiapp.translate_key({keyboard.shift}, 'A')

        assert [curses.KEY_SRIGHT] == cuiapp.translate_key(
            {keyboard.shift}, keyboard.right)

    def test_ctrl(self):
        cuiapp = app.CuiApp(kaa.options.build_parser().parse_args([]))
        assert '\x01' == cuiapp.translate_key({keyboard.ctrl}, 'a')
        assert '\x01' == cuiapp.translate_key({keyboard.ctrl}, 'A')

        with raises(KaaError):
            cuiapp.translate_key({keyboard.ctrl}, '-')

        assert [curses.KEY_SRIGHT] == cuiapp.translate_key(
            {keyboard.shift}, keyboard.right)

    def test_alt(self):
        cuiapp = app.CuiApp(kaa.options.build_parser().parse_args([]))
        assert '\x1ba' == cuiapp.translate_key({keyboard.alt}, 'a')
        assert '\x1b\x01' == cuiapp.translate_key(
            {keyboard.ctrl, keyboard.alt}, 'a')
        assert ['\x1b', curses.KEY_RIGHT] == cuiapp.translate_key(
            {keyboard.alt}, keyboard.right)
        assert ['\x1b', curses.KEY_SRIGHT] == cuiapp.translate_key(
            {keyboard.alt, keyboard.shift}, keyboard.right)



from kaa.filetype.default import defaultmode
from kaa.syntax_highlight import *

JSONThemes = {
    'basic': []
}


def make_tokenizer():
    return Tokenizer(tokens=[
        ('numeric', SingleToken('number',
                                [r'\b[0-9]+(\.[0-9]*)*\b', r'\b\.[0-9]+\b'])),
        ('string', Span('string', '"', '"', escape='\\'),)
    ])


class JSONMode(defaultmode.DefaultMode):
    MODENAME = 'JSON'
    tokenizer = make_tokenizer()

    def init_themes(self):
        super().init_themes()
        self.themes.append(JSONThemes)



import enchant
import kaa
from kaa import document
from kaa.ui.dialog import dialogmode
from kaa.ui.selectlist import filterlist
from kaa.keyboard import *
from kaa import doc_re
from kaa.ui.wordcomplete import wordcompletemode


class SpellCheckerWordInputMode(wordcompletemode.WordCompleteInputMode):
    MAX_INPUT_HEIGHT = 4
    INITIAL_MESSAGE = "Hit tab/shift+tab to complete."

    @classmethod
    def build(cls, target):
        doc = document.Document()
        mode = cls()
        doc.setmode(mode)
        mode.target = target

        with dialogmode.FormBuilder(doc) as f:
            f.append_text('caption', 'Select word:')
            f.append_text('default', ' ')
            f.append_text('default', '', mark_pair='query')
            f.append_text('default', ' ')

            f.append_text('right-button', '[&Replace]',
                          shortcut_style='right-button.shortcut',
                          on_shortcut=mode.selected)

            f.append_text('right-button', '[&Add word]',
                          shortcut_style='right-button.shortcut',
                          on_shortcut=mode.register)

            f.append_text('right-button', '[&Skip]',
                          shortcut_style='right-button.shortcut',
                          on_shortcut=mode.skip)

            f.append_text('right-button', '[&Quit]',
                          shortcut_style='right-button.shortcut',
                          on_shortcut=mode.quit)

        return doc

    def on_edited(self, wnd):
        filterlist.FilterListInputDlgMode.on_edited(self, wnd)

    def register(self, wnd):
        self.pwl.add(self.word)
        if self.callback:
            self.callback(self.wordpos[1])

        wnd.get_label('popup').destroy()

    def skip(self, wnd):
        if self.callback:
            self.callback(self.wordpos[1])

        wnd.get_label('popup').destroy()

    def quit(self, wnd):
        wnd.get_label('popup').destroy()

    def start(self, wnd, list, f, t, word, pwl, sugest):
        self.orgloc = self.target.get_cursor_loc()
        self.orgpos = self.target.cursor.pos
        self.wordpos = (f, t)
        self.word = word
        self.pwl = pwl

        list.document.mode.set_candidates(sugest)
        list.document.mode.set_query(list, '')

        list.get_label('popup').on_console_resized()


def show_suggests(wnd, f, t, s, pwl, sugest, callback):
    # select current word
    wnd.activate()
    wnd.cursor.setpos(t, middle=True)
    wnd.screen.selection.set_range(f, t)
    wnd.update_window()

    # todo: refactor
    doc = SpellCheckerWordInputMode.build(wnd)
    doc.mode.orgloc = wnd.get_cursor_loc()
    doc.mode.callback = callback

    if doc.mode.orgloc[0] > (wnd.mainframe.height // 2):
        doc.mode.stack_upper = True
    else:
        doc.mode.stack_upper = False

    dlg = kaa.app.show_dialog(doc)

    filterlistdoc = wordcompletemode.WordCompleteList.build()
    filterlistdoc.mode.SEP = ' '
    list = dlg.add_doc('dlg_filterlist', 0, filterlistdoc)

    doc.mode.start(wnd, list, f, t, s, pwl, sugest)

    return dlg


RE_WORD = doc_re.compile(r"\b[a-zA-Z][a-z']{2,}\b", doc_re.A)


def iter_words(wnd, pos):
    d = enchant.DictWithPWL("en_US", kaa.app.config.spellchecker_pwl)

    while True:
        posto = wnd.document.marks['spellchecker'][1]
        m = RE_WORD.search(wnd.document, pos, posto)
        if not m:
            return

        f, t = m.span()
        s = wnd.document.gettext(f, t)
        if d.check(s):
            pos = t
            continue

        suggest = d.pwl.suggest(s)
        if not suggest:
            pos = t
            continue

        pos = yield (f, t, s, d, suggest)


def run_spellchecker(wnd):
    sel = wnd.screen.selection.get_selrange()
    if sel:
        pos, posto = sel
    else:
        pos, posto = 0, wnd.document.endpos()

    wnd.document.marks['spellchecker'] = (pos, posto)

    wnd.screen.selection.clear()
    words = iter_words(wnd, pos)

    def callback(new_pos):
        if new_pos is None:
            # canceled
            del wnd.document.marks['spellchecker']
            return

        try:
            f, t, s, d, sugest = words.send(new_pos)
            show_suggests(wnd, f, t, s, d, sugest, callback)
        except StopIteration:
            del wnd.document.marks['spellchecker']

    rec = next(words, None)
    if rec:
        f, t, s, d, sugest = rec
        show_suggests(wnd, f, t, s, d, sugest, callback)



import kaa
import codecs

encodings = [
    'ascii',
    'big5',
    'big5hkscs',
    'cp037',
    'cp424',
    'cp437',
    'cp500',
    'cp720',
    'cp737',
    'cp775',
    'cp850',
    'cp852',
    'cp855',
    'cp856',
    'cp857',
    'cp858',
    'cp860',
    'cp861',
    'cp862',
    'cp863',
    'cp864',
    'cp865',
    'cp866',
    'cp869',
    'cp874',
    'cp875',
    'cp932',
    'cp949',
    'cp950',
    'cp1006',
    'cp1026',
    'cp1140',
    'cp1250',
    'cp1251',
    'cp1252',
    'cp1253',
    'cp1254',
    'cp1255',
    'cp1256',
    'cp1257',
    'cp1258',
    'euc-jp',
    'euc-jis-2004',
    'euc-jisx0213',
    'euc-kr',
    'gb2312',
    'gbk',
    'gb18030',
    'hz',
    'iso-2022-jp',
    'iso-2022-jp-1',
    'iso-2022-jp-2',
    'iso-2022-jp-2004',
    'iso-2022-jp-3',
    'iso-2022-jp-ext',
    'iso-2022-kr',
    'latin-1',
    'iso-8859-2',
    'iso-8859-3',
    'iso-8859-4',
    'iso-8859-5',
    'iso-8859-6',
    'iso-8859-7',
    'iso-8859-8',
    'iso-8859-9',
    'iso-8859-10',
    'iso-8859-13',
    'iso-8859-14',
    'iso-8859-15',
    'iso-8859-16',
    'johab',
    'koi8-r',
    'koi8-u',
    'mac-cyrillic',
    'mac-greek',
    'mac-iceland',
    'mac-latin2',
    'mac-roman',
    'mac-turkish',
    'ptcp154',
    'shift_jis',
    'shift_jis-2004',
    'shift_jisx0213',
    'utf-32',
    'utf-32be',
    'utf-32le',
    'utf-16',
    'utf-16be',
    'utf-16le',
    'utf-7',
    'utf-8',
    'utf-8-sig',
]

canonical_names = {codecs.lookup(name).name: name for name in encodings}


# from tokenize.py
def _get_normal_name(orig_enc):
    """Imitates get_normal_name in tokenizer.c."""
    # Only care about the first 12 characters.
    enc = orig_enc[:12].lower().replace("_", "-")
    if enc == "utf-8" or enc.startswith("utf-8-"):
        return "utf-8"
    if enc in ("latin-1", "iso-8859-1", "iso-latin-1") or \
       enc.startswith(("latin-1-", "iso-8859-1-", "iso-latin-1-")):
        return "iso-8859-1"
    return orig_enc


def normalize_encname(name, default='utf-8'):
    name = _get_normal_name(name)
    try:
        name = codecs.lookup(name).name
    except LookupError:
        kaa.log.error('Unknown encoding', exc_info=True)
        name = default   # error fallback
    return canonical_names[name]



import os
import sys
import subprocess
import shutil
import kaa
import kaa.log


def select_clipboard():
    """Select clipboard class for the platform."""

    if sys.platform == 'darwin':
        return MacClipboard()
    elif X11Clipboard.check():
        return X11Clipboard()
    else:
        return Clipboard()


class Clipboard:

    """Basic clipboard class without platform's native clipboard."""

    # Max history of clipboard entry.
    MAX_CLIPBOARD = 10

    def _get_hist(self):
        return kaa.app.config.hist('clipboard', max_hist=self.MAX_CLIPBOARD)

    def get(self):
        """Get current clipboard entry."""

        all = self.get_all()
        ret = all[0] if all else ''
        if ret:
            self._get_hist().add(ret)
        return ret

    def get_all(self):
        """Get clipboard history."""

        return [s for s, i in self._get_hist().get() if s]

    def _set(self, s):
        self._get_hist().add(s)

    def set(self, s):
        self._set(s)


class NativeClipboard(Clipboard):

    """Base class of platform's native clipboard."""

    def get(self):
        try:
            s = self._get_native_clipboard()
            if s:
                self._set(s)
                return s
            else:
                return super().get()

        except Exception:
            kaa.log.error('Error to copy', exc_info=True)
            return super().get()

    def set(self, s):
        super().set(s)
        try:
            self._set_native_clipboard(s)
        except Exception:
            kaa.log.error('Error to paste', exc_info=True)


class MacClipboard(NativeClipboard):

    """For MAC OS X"""

    COPYCOMMAND = 'pbcopy'
    PASTECOMMAND = 'pbpaste'

    def _get_native_clipboard(self):
        return subprocess.check_output(self.PASTECOMMAND, shell=True,
                                       universal_newlines=True)

    def _set_native_clipboard(self, s):
        p = subprocess.Popen(self.COPYCOMMAND, stdin=subprocess.PIPE,
                             universal_newlines=True, shell=True)
        p.communicate(s)
        p.wait()


class X11Clipboard(NativeClipboard):

    """For UN*X"""

    CLIPCOMMAND = 'xclip'
    COPYCOMMAND = 'xclip -i -selection clipboard'
    PASTECOMMAND = 'xclip  -o -selection clipboard'

    @classmethod
    def check(cls):
        if not shutil.which(X11Clipboard.CLIPCOMMAND):
            return False

        if 'DISPLAY' not in os.environ:
            return False

        return True

    def _get_native_clipboard(self):
        return subprocess.check_output(self.PASTECOMMAND, shell=True,
                                       universal_newlines=True)

    def _set_native_clipboard(self, s):
        p = subprocess.Popen(
            self.COPYCOMMAND,
            stdin=subprocess.PIPE,
            shell=True,
            universal_newlines=True)
        p.stdin.write(s)
        p.stdin.close()
        p.wait()



import os
import importlib
import sqlite3
import json
import kaa
from kaa import consts, utils


class KaaHistoryStorage:
    # todo: lock directory for network drives?

    def __init__(self, filename):
        self.filename = filename
        self.conn = sqlite3.connect(filename)
        self.conn.execute('''
            PRAGMA synchronous = NORMAL
        ''')
        self.hists = {}
        self._updated = False

    BUFFER_DURATION = 10

    def updated(self):
        if not self._updated:
            kaa.app.call_later(self.BUFFER_DURATION, self.flush)
            self._updated = True

    def flush(self):
        try:
            if self._updated:
                for hist in self.hists.values():
                    hist.flush()
                self.conn.commit()
                self._updated = False
        except Exception:
            kaa.log.error('Error saving history', exc_info=True)

    def close(self):
        for hist in self.hists.values():
            hist.flush()

        for hist in self.hists.values():
            hist.close()

        self.conn.commit()
        self.conn.close()
        self.hists = None

    def get_history(self, name, max_hist=None):
        if name not in self.hists:
            hist = History(self, name, max_hist)
            self.hists[name] = hist

        return self.hists[name]


class History:
    MAX_HISTORY = 250

    def __init__(self, storage, name, max_history):
        self.storage = storage
        self.name = name
        self.table_name = 'hist_' + self.name

        self._register()
        self._max_history = max_history
        if not self._max_history:
            self._max_history = self.MAX_HISTORY

        self.buffer = []

    def _register(self):
        self.storage.conn.execute('''
            CREATE TABLE IF NOT EXISTS {}
                (id INTEGER PRIMARY KEY,
                value TEXT,
                info  TEXT)'''.format(self.table_name))

    @utils.ignore_errors
    def close(self):
        ret = self.storage.conn.execute('''
            SELECT id FROM {} ORDER BY id DESC LIMIT ?
            '''.format(self.table_name), (self._max_history * 2, ))
        recs = [value for value in ret]
        if recs:
            last, = recs[-1]
            self.storage.conn.execute('''
                DELETE FROM {} WHERE id < ?
                '''.format(self.table_name), (last,))

    @utils.ignore_errors
    def add(self, value, info=None):
        for n, (buf_value, buf_info) in enumerate(self.buffer):
            if value == buf_value:
                del self.buffer[n]
                break
        self.buffer.insert(0, (value, info))
        self.storage.updated()

    def get(self):
        buf_values = set(value for value, info in self.buffer)

        ret = self.storage.conn.execute('''
            SELECT value, info FROM {} ORDER BY id DESC LIMIT ?
            '''.format(self.table_name),
            (self._max_history,))

        return self.buffer + [(value, json.loads(info))
                              for value, info in ret if value not in buf_values]

    @utils.ignore_errors
    def find(self, value):
        for buf_value, buf_info in self.buffer:
            if value == buf_value:
                return buf_info

        ret = self.storage.conn.execute('''
            SELECT info FROM {} WHERE value=?
            '''.format(self.table_name),
            (value,))

        for info, in ret:
            return json.loads(info)

    @utils.ignore_errors
    def flush(self):
        # Supress db write if record don't changed.
        if len(self.buffer) == 1:
            ret = self.storage.conn.execute('''
                SELECT value, info FROM {0} as T
                    WHERE T.id = (SELECT max(id) FROM {0})
                '''.format(self.table_name))
            rec = ret.fetchone()
            if rec:
                if self.buffer[0][0] == rec[0]:
                    if self.buffer[0][1] == json.loads(rec[1]):
                        del self.buffer[:]
                        return

        for value, info in self.buffer[::-1]:
            ret = self.storage.conn.execute('''
                DELETE FROM {} WHERE value = ?'''.format(self.table_name),
                                            (value, ))

            self.storage.conn.execute('''
                INSERT INTO {}(value, info) VALUES(?, ?)
                '''.format(self.table_name), (value, json.dumps(info)))
        self.buffer = []


class Config:
    FILETYPES = [
        'kaa.filetype.default',
        'kaa.filetype.python',
        'kaa.filetype.html',
        'kaa.filetype.javascript',
        'kaa.filetype.css',
        'kaa.filetype.diff',
        'kaa.filetype.rst',
        'kaa.filetype.markdown',
        'kaa.filetype.c',
        'kaa.filetype.ini',
        'kaa.filetype.json',
    ]

    DEFAULT_NEWLINE = 'auto'
    DEFAULT_ENCODING = 'utf-8'
    AMBIGUOUS_WIDTH = 1

    def __init__(self, option):
        kaadir = os.path.abspath(
            os.path.expandvars(
                os.path.expanduser(consts.KAA_DIR)))

        if not os.path.exists(kaadir):
            os.makedirs(kaadir)

        logdir = os.path.join(kaadir, consts.LOGDIR)
        if not os.path.exists(logdir):
            os.mkdir(logdir)

        histdir = os.path.join(kaadir, consts.HISTDIR)
        if not os.path.exists(histdir):
            os.mkdir(histdir)

        self.KAADIR = kaadir
        self.LOGDIR = logdir
        self.HISTDIR = histdir

        self.palette = option.palette
        self.spellchecker_pwl = os.path.join(kaadir, consts.SPELLCHECKER_PWL)

    def init_history(self):
        self.hist_storage = KaaHistoryStorage(
            os.path.join(self.HISTDIR, consts.HIST_DBNAME))

        self._config = self.hist_storage.get_history('Values')

    def close(self):
        self.hist_storage.close()

    def hist(self, name, max_hist=None):
        return self.hist_storage.get_history(name, max_hist=max_hist)

    def get_mode_packages(self):
        for pkgname in self.FILETYPES:
            try:
                pkg = importlib.import_module(pkgname)
            except Exception:
                kaa.log.exception('Error loading filetype: ' + repr(pkgname))
            else:
                yield pkg

    def save_value(self, name, value):
        self._config.add(name, value)

    def load_value(self, name, default=None):
        values = self._config.get()
        for n, info in values:
            if name == n:
                return info
        return default



import kaa_testutils
from kaa import syntax_highlight


class TestHighlight(kaa_testutils._TestDocBase):

    def test_token(self):
        doc = self._getdoc(' abc def ghi ')

        root = syntax_highlight.Tokenizer(tokens=[
            ('keyword', syntax_highlight.Keywords('keyword', ['abc', 'ghi']))
        ])

        tokenizer = syntax_highlight.begin_tokenizer(
            doc, root, 0)

        styles = list(tokenizer)
        assert styles == [
            (0, 1, root.styleid_default),
            (1, 4, root.tokens.keyword.styleid_token),
            (4, 9, root.styleid_default),
            (9, 12, root.tokens.keyword.styleid_token),
            (12, 13, root.styleid_default), ]

    def test_span(self):
        doc = self._getdoc(' ab\\cc def ghi ')

        root = syntax_highlight.Tokenizer(tokens=[
            ('span1', syntax_highlight.Span('span', 'a', 'c',
                                            escape='\\')),
            ('span2', syntax_highlight.Span('span', 'd', 'f',
                                            capture_end=False)),
            ('span3', syntax_highlight.Span('span', 'g', 'i', terminate_tokens='h')),
        ])

        tokenizer = syntax_highlight.begin_tokenizer(
            doc, root, 0)

        styles = list(tokenizer)
        assert styles == [
            (0, 1, root.styleid_default),
            (1, 2, root.tokens.span1.styleid_span),
            (2, 5, root.tokens.span1.styleid_span),
            (5, 6, root.tokens.span1.styleid_span),
            (6, 7, root.styleid_default),
            (7, 8, root.tokens.span2.styleid_span),
            (8, 9, root.tokens.span2.styleid_span),
            (9, 11, root.styleid_default),
            (11, 12, root.tokens.span3.styleid_span),
        ]

    def test_resume(self):
        doc = self._getdoc('    abc')

        root = syntax_highlight.Tokenizer(tokens=[
            ('keyword', syntax_highlight.Keywords('keyword', ['abc', 'ghi']))])

        tokeniter = syntax_highlight.begin_tokenizer(
            doc, root, 2)

        styles = list(tokeniter)
        assert styles == [
            (0, 4, root.styleid_default),
            (4, 7, root.tokens.keyword.styleid_token)]

        tokeniter = syntax_highlight.begin_tokenizer(
            doc, root, 6)

        doc.styles.setints(4, 7, root.tokens.keyword.styleid_token)
        styles = list(tokeniter)
        assert styles == [
            (0, 4, root.styleid_default),
            (4, 7, root.tokens.keyword.styleid_token)]


class TestHighlight_Nest(kaa_testutils._TestDocBase):

    def test_nest_resume(self):

        doc = self._getdoc('   abc   123 x abc')

        class SubKeyword(syntax_highlight.Keywords):

            def on_start(self, doc, match):
                pos, terminates = yield from super().on_start(doc, match)
                return (yield from sub.run(doc, pos)), False

        root = syntax_highlight.Tokenizer(tokens=[
            ('subkeyword', SubKeyword('keyword', ['abc']))])

        sub = syntax_highlight.Tokenizer(parent=root, terminates='x', tokens=[
            ('keyword', syntax_highlight.Keywords('keyword', ['123']))])

        doc.styles.setints(0, 3, root.styleid_default)
        doc.styles.setints(3, 6, root.tokens.subkeyword.styleid_token)
        doc.styles.setints(6, 9, sub.styleid_default)
        doc.styles.setints(9, 12, sub.tokens.keyword.styleid_token)

        # parse at blank
        tokeniter = syntax_highlight.begin_tokenizer(
            doc, root, 2)

        styles = list(tokeniter)
        assert styles == [
            (0, 3, root.styleid_default),
            (3, 6, root.tokens.subkeyword.styleid_token),
            (6, 9, sub.styleid_default),
            (9, 12, sub.tokens.keyword.styleid_token),
            (12, 13, sub.styleid_default),
            (13, 15, root.styleid_default),
            (15, 18, root.tokens.subkeyword.styleid_token)]

        # parse at first token
        tokeniter = syntax_highlight.begin_tokenizer(
            doc, root, 4)

        styles = list(tokeniter)
        assert styles == [
            (0, 3, root.styleid_default),
            (3, 6, root.tokens.subkeyword.styleid_token),
            (6, 9, sub.styleid_default),
            (9, 12, sub.tokens.keyword.styleid_token),
            (12, 13, sub.styleid_default),
            (13, 15, root.styleid_default),
            (15, 18, root.tokens.subkeyword.styleid_token)]

        # parse at blank of sub2
        tokeniter = syntax_highlight.begin_tokenizer(
            doc, root, 8)

        styles = list(tokeniter)
        assert styles == [
            (6, 9, sub.styleid_default),
            (9, 12, sub.tokens.keyword.styleid_token),
            (12, 13, sub.styleid_default),
            (13, 15, root.styleid_default),
            (15, 18, root.tokens.subkeyword.styleid_token)]

        # parse at blank of keyword2
        tokeniter = syntax_highlight.begin_tokenizer(
            doc, root, 11)

        styles = list(tokeniter)
        assert styles == [
            (6, 9, sub.styleid_default),
            (9, 12, sub.tokens.keyword.styleid_token),
            (12, 13, sub.styleid_default),
            (13, 15, root.styleid_default),
            (15, 18, root.tokens.subkeyword.styleid_token)]



import copy

from kaa import doc_re
from kaa.command import commandid, norec, norerun
from kaa.filetype.default import defaultmode
from kaa.keyboard import ctrl
from kaa.theme import Style
from kaa.syntax_highlight import *

INIThemes = {
    'basic': [
        Style('section', 'Blue', None, bold=True),
        Style('param-name', 'Orange', None, bold=True),
    ]
}


def ini_tokens():
    return [
        ('comment', Span('comment', r';', '$')),
        ('section', Span('section', r'^\[', r'\]')),
        ('param-name', SingleToken('param-name', [r"^([a-zA-Z0-9_-])+"])),
    ]


def make_tokenizer():
    return Tokenizer(tokens=ini_tokens())

INIMENU = [
    ['&Comment', None, 'code.region.linecomment'],
    ['&Uncomment', None, 'code.region.unlinecomment'],
    ['&Table of contents', None, 'toc.showlist'],
]

MDMENU = [
    ['&Table of contents', None, 'toc.showlist'],
]

ini_keys = {
    (ctrl, 't'): 'toc.showlist',
}


class INIMode(defaultmode.DefaultMode):
    MODENAME = 'INI'
    LINE_COMMENT = ';'

    RE_SECTION = doc_re.compile(r'^\[(?P<SECTION>.+)\]$', doc_re.M)

    tokenizer = make_tokenizer()

    def init_keybind(self):
        super().init_keybind()
        self.register_keys(self.keybind, [ini_keys])

    def init_themes(self):
        super().init_themes()
        self.themes.append(INIThemes)

    def init_menu(self):
        super().init_menu()
        self.menu['CODE'] = copy.deepcopy(INIMENU)

    def get_sections(self):
        pos = 0
        while True:
            m = self.RE_SECTION.search(self.document, pos)
            if not m:
                break

            pos = m.end()
            name = m.group('SECTION')

            yield self.HeaderInfo('namespace', None, name, name,
                                  None, m.start())

    @commandid('toc.showlist')
    @norec
    @norerun
    def show_toclist(self, wnd):
        from kaa.ui.toclist import toclistmode
        sections = list(self.get_sections())
        toclistmode.show_toclist(wnd, sections)



import copy
from kaa.filetype.default import defaultmode
from kaa import doc_re
from kaa.theme import Style
from kaa.command import commandid
from kaa.command import norec
from kaa.command import norerun
from kaa.keyboard import *
from kaa.syntax_highlight import *


MarkdownThemes = {
    'basic': [
        Style('escape', 'default', 'default'),
        Style('header', 'Blue', None),
        Style('hr', 'Green', None),
        Style('strong', 'Magenta', None),
        Style('emphasis', 'Blue', None),
        Style('literal', 'Cyan', None),
        Style('reference', 'Red', None),
        Style('role', 'Cyan', None),
        Style('substitution', 'Green', None),
    ],
}


class LinkToken(Span):
    # [xxx](yyy "xzzzz")
    # [xxx]:
    # ![xxx]

    def __init__(self, stylename):
        super().__init__(stylename, r'!?\[', r'][:\(]?', escape=r'\\')

    def on_start(self, doc, match):
        pos, terminates = yield from super().on_start(doc, match)
        if pos == doc.endpos():
            return pos, terminates
        c = doc.gettext(pos - 1, pos)

        if c == '(':
            # [xxx](yyy "xzzzz")
            pos = yield from self.tokenizer._LinkTokenizer.run(doc, pos)

        return pos, False


def is_list_block(doc, pos, tokenizer):
    '''Returns True if pos is in list block'''

    pos = doc.gettol(pos)
    while pos:
        pos -= 1
        tol = doc.gettol(pos)
        s = doc.gettext(tol, pos).rstrip()
        if not s:
            # skip blank line
            continue
        token = tokenizer.get_token_at(doc, tol)
        if token is tokenizer.tokens.list:
            return True

        if s[0] not in ['\t ']:  # check previous line if indended.
            break
        pos = tol


class List(SingleToken):

    def on_start(self, doc, match):
        if match.group('list_indent'):
            # has indent
            f, t = match.span()
            if not is_list_block(doc, f, self.tokenizer):
                t = doc.geteol(f)
                yield (f, t, self.tokenizer.tokens.code1.styleid_span)
                return t, self.terminates

        ret = yield from super().on_start(doc, match)
        return ret


class PreBlock(Span):

    def on_start(self, doc, match):
        # skip if line is nested list block
        #
        # 1. abcdefg
        #     2. hijklmn

        if is_list_block(doc, match.start(), self.tokenizer):
            f, t = match.span(1)
            yield (f, t, self.tokenizer.styleid_default)
            return t, self.terminates

        ret = yield from super().on_start(doc, match)
        return ret


def make_tokenizer():
    ret = Tokenizer(tokens=(
        ('escape', SingleToken('escape', [r'\\.'])),
        ('hr', SingleToken('hr', [r'^(\-{3,}|_{3,}|\*{3,})$'])),

        ('header1', SingleToken('header', [r'^.+\n(?P<H1>[-=])(?P=H1)+$'])),
        ('header2', SingleToken('header', [r'^\#{1,6}.*$'])),

        ('strong1', Span('strong', r'\*\*(?!\s)', r'\*\*|$', escape='\\')),
        ('strong2', Span('strong', r'__(?!\s)', r'__|$', escape='\\')),

        ('emphasis1', Span('emphasis', r'\*(?!\s)', r'\*|$', escape='\\')),
        ('emphasis2', Span('emphasis', r'_(?!\s)', r'_|$', escape='\\')),

        ('list', List('default', [r'^(?P<list_indent>\s*)(\*\ |\d+\.\ )'])),
        ('code1', PreBlock('literal', r'^(\ {4,})', r'$')),
        ('code2', Span('literal', r'^```', r'^```\s*$', escape='\\')),
        ('code3', Span('literal', r'`(?!\s)', r'`|$', escape='\\')),

        ('link', LinkToken('reference')),
    ))

    ret._LinkTokenizer = Tokenizer(parent=ret,
                                   default_style='reference',
                                   tokens=(
                                       ('desc',
                                        Span('reference', r'"', r'"', escape='\\')),
                                       ('close',
                                        SingleToken(
                                            'reference', [r"\)"], terminates=True)),
                                   ))
    return ret


MDMENU = [
    ['&Table of contents', None, 'toc.showlist'],
]

md_keys = {
    (ctrl, 't'): 'toc.showlist',
}


class MarkdownMode(defaultmode.DefaultMode):
    MODENAME = 'Markdown'
    tokenizer = make_tokenizer()

    def init_keybind(self):
        super().init_keybind()
        self.register_keys(self.keybind, [md_keys])

    def init_menu(self):
        super().init_menu()
        self.menu['CODE'] = copy.deepcopy(MDMENU)

    def init_themes(self):
        super().init_themes()
        self.themes.append(MarkdownThemes)

    def init_tokenizer(self):
        self.tokenizer = MarkdownTokenizer

    HEADER1 = r'^(?P<TITLE>.+)\n(?P<H1>[=-])(?P=H1)+$'
    HEADER2 = r'^(?P<H2>\#{1,6})(?P<TITLE2>.+)$'

    RE_HEADER = doc_re.compile(
        '|'.join([HEADER1, HEADER2]), doc_re.X | doc_re.M)

    def get_headers(self):
        stack = []
        pos = 0
        while True:
            m = self.RE_HEADER.search(self.document, pos)
            if not m:
                break

            pos = m.end()
            name = m.group('TITLE') or m.group('TITLE2')
            name = name.strip()

            bar = m.group('H1')
            if bar:
                if bar == '=':
                    level = 0
                else:
                    level = 1
            else:
                level = len(m.group('H2')) - 1

            if not stack:
                header = self.HeaderInfo('namespace', None, name, name,
                                         None, m.start())
                yield header
                stack.append((level, header))
                continue

            for i, (parent_level, info) in enumerate(stack):
                if level <= parent_level:
                    del stack[i:]
                    break

            if stack:
                parent = stack[-1][1]
            else:
                parent = None

            header = self.HeaderInfo(
                'namespace', parent, name,
                name, None, m.start())
            yield header
            stack.append((level, header))

    @commandid('toc.showlist')
    @norec
    @norerun
    def show_toclist(self, wnd):
        from kaa.ui.toclist import toclistmode
        headers = list(self.get_headers())
        toclistmode.show_toclist(wnd, headers)



import re
import itertools
import unicodedata
import collections

from kaa import doc_re
import kaa
from kaa import keyboard, editmode
import kaa.log
from kaa import theme
from kaa import screen
from kaa import addon
from kaa import syntax_highlight


class SearchOption:
    RE = doc_re

    def __init__(self):
        self.text = ''
        self.replace_to = ''
        self.ignorecase = True
        self.word = False
        self.regex = False

    def clone(self):
        ret = self.__class__()
        ret.text = self.text
        ret.replace_to = self.replace_to
        ret.ignorecase = self.ignorecase
        ret.word = self.word
        ret.regex = self.regex
        return ret

    def get_regex(self):
        text = self.text
        if not self.regex:
            text = self.RE.escape(text)
        if self.word:
            text = r'\b' + text + r'\b'

        opt = self.RE.MULTILINE
        if self.ignorecase:
            opt += self.RE.IGNORECASE

        regex = self.RE.compile(text, opt)
        return regex

SearchOption.LAST_SEARCH = SearchOption()

DefaultTokenizer = syntax_highlight.Tokenizer()


class ModeBase:
    (UNDO_INSERT,
     UNDO_REPLACE,
     UNDO_DELETE) = range(3)

    CLOSE_ON_DEL_WINDOW = True

    SCREEN_NOWRAP = False
    NO_WRAPINDENT = False
    SHOW_LINENO = False
    SHOW_BLANK_LINE = False
    USE_UNDO = True
    DOCUMENT_MODE = False
    HIGHLIGHT_CURSOR_ROW = False
    SHOW_CURSOR = True
    DELAY_STR = True

    tab_width = 8
    indent_width = 4
    indent_tab = False
    auto_indent = True

    closed = False
    theme = None
    _check_fileupdate = 0
    _last_autoindent = None

    tokenizer = DefaultTokenizer

    DEFAULT_MENU_MESSAGE = ''

    @classmethod
    def update_fileinfo(cls, fileinfo, document=None):
        pass

    def __init__(self):
        self.commands = {}
        self.is_availables = {}

        self.keybind = keyboard.KeyBind()
        self.keybind_vi_commandmode = keyboard.KeyBind()
        self.keybind_vi_visualmode = keyboard.KeyBind()
        self.keybind_vi_visuallinewisemode = keyboard.KeyBind()
        self.init_keybind()
        self.init_addon_keys()

        self.init_commands()
        self.init_addon_commands()

        self.init_menu()

        self.themes = []
        self.init_themes()
        self.init_addon_themes()

        self.tokenizers = []
        self.init_tokenizers()

        self._highlight_done = 0
        self._highlight_iter = None

        self.stylemap = {}
        self.stylenamemap = {}

        self.setup()
        self.setup_addon()

        self._build_theme()
        kaa.app.translate_theme(self.theme)

    def setup_addon(self):
        names = []
        for cls in self.__class__.__mro__:
            modulename = cls.__module__
            name = '.'.join((modulename, cls.__name__))
            names.insert(0, name)

        for name in names:
            for f in addon.get_addon(name, 'setup'):
                f(self)

    def setup(self):
        pass

    def close(self):
        self.document = None
        self.closed = True

        self.keybind.clear()

        self.theme = None

        self.commands.clear()
        self.commands = None

        self.is_availables = None
        self.tokenizers = None
        self.tokenizer = None
        self.stylemap = None
        self._highlight_done = 0
        self._highlight_iter = None

    def _build_style_map(self):
        self.stylemap[0] = self.theme.get_style('default')
        for styleid, token in self.tokenizer.styleid_map.items():
            if token:
                stylename = token.get_stylename(styleid)
                style = self.theme.get_style(stylename)

                self.stylemap[styleid] = style
                self.stylenamemap[style.name] = styleid
            else:
                style = self.theme.get_style('default')
                self.stylemap[styleid] = style
                self.stylenamemap[style.name] = styleid

    def on_set_document(self, document):
        self.document = document
        self.document.styles.setints(0, len(self.document.styles), 0)

        self._build_style_map()
        self.document.use_undo(self.USE_UNDO)

    def on_document_updated(self, pos, inslen, dellen):
        self._last_autoindent = None
        if pos <= self._highlight_done:
            self._highlight_done = pos
            self._highlight_iter = None

    def on_file_saved(self, fileinfo):
        pass

    def on_add_window(self, wnd):
        self.editmode_insert(wnd)

    def on_del_window(self, wnd):
        if len(self.document.wnds) == 1:
            if self.DOCUMENT_MODE and self.document.fileinfo:
                self.document.fileinfo.save_screenloc(wnd.screen.pos)

    def on_focus(self, wnd):
        kaa.app.show_cursor(self.is_cursor_visible())

        # relocate cursor
        wnd.cursor.setpos(wnd.cursor.pos)

    def register_keys(self, keybind, keys):
        for d in keys:
            keybind.add_keybind(d)

    def init_keybind(self):
        pass

    def get_class_name(self):
        modulename = self.__class__.__module__
        name = '.'.join((modulename, self.__class__.__name__))
        return name

    def add_keybinds(self, editmode='input', keys=None):
        if not keys:
            return
        if editmode == 'input':
            self.keybind.add_keybind(keys)
        elif editmode == 'command':
            self.keybind_vi_commandmode.add_keybind(keys)
        elif editmode == 'visual':
            self.keybind_vi_visualmode.add_keybind(keys)
        elif editmode == 'visualline':
            self.keybind_vi_visuallinewisemode.add_keybind(keys)

    def init_addon_keys(self):
        name = self.get_class_name()
        keydef = addon.get_addon(name, 'keybind')
        for edit_mode, keys in keydef:
            if edit_mode == 'input':
                self.keybind.add_keybind(keys)
            elif edit_mode == 'command':
                self.keybind_vi_commandmode.add_keybind(keys)
            elif edit_mode == 'visual':
                self.keybind_vi_visualmode.add_keybind(keys)
            elif edit_mode == 'visualline':
                self.keybind_vi_visuallinewisemode.add_keybind(keys)

    def init_commands(self):
        from kaa.commands import (editorcommand, editmodecommand, vicommand)

        # todo: re-design command injection
        self.register_commandobj(editorcommand.CursorCommands())
        self.register_commandobj(editorcommand.EditCommands())
        self.register_commandobj(editorcommand.CodeCommands())
        self.register_commandobj(editorcommand.SelectionCommands())
        self.register_commandobj(editorcommand.SearchCommands())
        self.register_commandobj(editmodecommand.EditModeCommands())
        self.register_commandobj(vicommand.ViCommands())

        for name in dir(self):
            attr = getattr(self, name)
            if hasattr(attr, 'COMMAND_ID') and callable(attr):
                self.commands[getattr(attr, 'COMMAND_ID')] = attr

    def init_addon_commands(self):
        name = self.get_class_name()
        commands = addon.get_addon(name, 'command')

        for f in commands:
            self.commands[f.COMMAND_ID] = f

    def init_menu(self):
        pass

    def init_themes(self):
        pass

    def init_addon_themes(self):
        name = self.get_class_name()
        themes = addon.get_addon(name, 'style')
        self.themes.extend(themes)

    def add_theme(self, theme):
        self.themes.append(theme)

    def init_tokenizers(self):
        pass

    def add_command(self, f, command_id=None):
        if command_id is not None:
            f.COMMAND_ID = command_id

        self.commands[f.COMMAND_ID] = f

    def register_commandobj(self, cmds):
        self.commands.update(cmds.get_commands())
        self.is_availables.update(cmds.get_commands_is_enable())

    def get_command(self, commandid):
        ret = kaa.app.get_command(commandid)
        if ret:
            return ret
        is_available = self.is_availables.get(commandid, None)
        cmd = self.commands.get(commandid, None)
        return (is_available, cmd)

    def editmode_insert(self, wnd):
        wnd.set_editmode(editmode.EditMode())

    def editmode_replace(self, wnd):
        wnd.set_editmode(editmode.ReplaceMode())

    def editmode_visual(self, wnd):
        wnd.set_editmode(editmode.VisualMode())

    def editmode_visual_linewise(self, wnd):
        wnd.set_editmode(editmode.VisualLinewiseMode())

    def editmode_command(self, wnd):
        wnd.set_editmode(editmode.CommandMode())

    def get_styleid(self, stylename):
        if stylename == 'default':
            return 0

        styleid = self.stylenamemap.get(stylename, None)
        if styleid is not None:
            return styleid

        if not self.stylemap:
            ret = 1
        else:
            ret = max(self.stylemap.keys()) + 1

        style = self.theme.get_style(stylename)
        self.stylemap[ret] = style
        self.stylenamemap[stylename] = ret
        return ret

    def select_theme(self, theme_name, themes):
        theme = themes.get(theme_name, None)
        if theme is None:
            theme = themes[kaa.app.DEFAULT_THEME]
        return theme

    def _build_theme(self):
        theme_name = kaa.app.get_current_theme()
        self.theme = theme.Theme([])
        for t in self.themes:
            self.theme.update(self.select_theme(theme_name, t))
        self.theme.finish_update()

    def get_style(self, tokenid):
        ret = self.stylemap.get(tokenid, None)
        if ret is None:
            ret = self.theme.get_style('default')
        return ret

    def get_menu(self, itemname):
        return self.menu.get(itemname, None)

    def is_cursor_visible(self):
        return self.SHOW_CURSOR

    def on_keypressed(self, wnd, event, s, commands, candidate):
        return s, commands, candidate

    def filter_string(self, wnd, s):
        return s

    def on_edited(self, wnd):
        pass

    def on_str(self, wnd, s, overwrite=False):
        self.put_string(wnd, s, overwrite=overwrite)
        wnd.screen.selection.clear()

        if kaa.app.macro.is_recording():
            kaa.app.macro.record_string(s, overwrite)

        # run highlighter a bit to display changes immediately.
        self.run_tokenizer(batch=50)

    def on_commands(self, wnd, commandids, n_repeat=1):
        wnd.set_command_repeat(n_repeat)
        try:
            if callable(commandids):
                commandids(wnd)
                return

            current_lastcommands = []
            for commandid in commandids:
                is_available, command = self.get_command(commandid)
                if not command:
                    msg = 'command {!r} is not registered.'.format(commandid)
                    kaa.app.messagebar.set_message(msg)
                    kaa.log.error(msg)
                    return

                command(wnd)
                if kaa.app.macro.is_recording():
                    kaa.app.macro.record(n_repeat, command)

                if not getattr(command, 'NORERUN', False):
                    current_lastcommands.append(commandid)

            if current_lastcommands:
                kaa.app.lastcommands = (n_repeat, current_lastcommands)

        finally:
            wnd.set_command_repeat(1)

        def f():
            import kaa.cui.frame
            frames = kaa.cui.frame.MainFrame.childframes
            frame_panels = [f._panel for f in frames]

            import curses.panel
            panels = [curses.panel.top_panel()]
            while True:
                panel = panels[-1].below()
                if panel:
                    panels.append(panel)
                else:
                    break

            for panel in panels:
                if panel in frame_panels:
                    if panel is not frame_panels[0]:
                        _trace('Invalid frame order: %s%s' %
                               (self, commandids,))
                break

        f()

    def on_esc_pressed(self, wnd, event):
        pass

    def on_cursor_located(self, wnd, pos, y, x):
        pass

    def insert_string(self, wnd, pos, s, update_cursor=True):
        """Insert string"""

        cur_pos = wnd.cursor.pos

        wnd.document.insert(pos, s)

        if update_cursor:
            wnd.cursor.setpos(wnd.cursor.pos + len(s))
            wnd.cursor.savecol()

        if wnd.document.undo:
            wnd.document.undo.add(self.UNDO_INSERT, pos, s,
                                  cur_pos, wnd.cursor.pos)

        self.on_edited(wnd)

    def replace_string(self, wnd, pos, posto, s, update_cursor=True):
        """Replace string"""

        cur_pos = wnd.cursor.pos

        deled = wnd.document.gettext(pos, posto)
        wnd.document.replace(pos, posto, s)

        if update_cursor:
            wnd.cursor.setpos(pos + len(s))
            wnd.cursor.savecol()

        if wnd.document.undo:
            wnd.document.undo.add(self.UNDO_REPLACE, pos, posto, s,
                                  deled, cur_pos, wnd.cursor.pos)

        self.on_edited(wnd)

    def delete_string(self, wnd, pos, posto, update_cursor=True):
        """Delete string"""

        cur_pos = wnd.cursor.pos

        if pos < posto:
            deled = wnd.document.gettext(pos, posto)
            wnd.document.delete(pos, posto)

            if update_cursor:
                wnd.cursor.setpos(pos)
                wnd.cursor.savecol()

            if wnd.document.undo:
                wnd.document.undo.add(self.UNDO_DELETE, pos, posto, deled,
                                      cur_pos, wnd.cursor.pos)
            self.on_edited(wnd)

    def delete_ws(self, wnd, pos, update_cursor=False):
        m = doc_re.compile(r'[ \t]+').match(self.document, pos)
        if m:
            f, t = m.span()
            self.delete_string(wnd, f, t, update_cursor=update_cursor)
            return t - f

    def replace_rect(self, wnd, repto):
        with wnd.document.undo_group():
            (posfrom, posto, colfrom, colto
             ) = wnd.screen.selection.get_rect_range()

            for s in repto:
                if posto <= posfrom:
                    break

                sel = wnd.screen.selection.get_col_string(
                    posfrom, colfrom, colto)
                if sel:
                    f, t, org = sel
                    if org.endswith('\n'):
                        t = max(f, t - 1)
                    self.replace_string(wnd, f, t, s)
                    posto += (len(s) - (t - f))
                posfrom = wnd.document.geteol(posfrom)

    def put_string(self, wnd, s, overwrite=False):
        s = self.filter_string(wnd, s)

        if wnd.screen.selection.is_selected():
            if wnd.screen.selection.is_rectangular():
                if '\n' not in s:
                    self.replace_rect(wnd, itertools.repeat(s))
                else:
                    self.replace_rect(wnd, s.split('\n'))

            else:
                sel = wnd.screen.selection.get_selrange()
                f, t = sel
                self.replace_string(wnd, f, t, s)
        else:
            if not overwrite:
                self.insert_string(wnd, wnd.cursor.pos, s)
            else:
                eol = wnd.document.get_line_to(wnd.cursor.pos)
                posto = min(wnd.cursor.pos + len(s), eol)
                self.replace_string(wnd, wnd.cursor.pos, posto, s)

    def update_charattr(self, wnd):
        if wnd.charattrs:
            wnd.charattrs.clear()
            wnd.screen.style_updated()
            return True

    def on_idle(self):
        if self.closed:
            return

        ret = self.run_tokenizer()
        return ret

    def get_line_overlays(self):
        return {}

    def _get_highlight_range(self):
        return (0, self.document.endpos())

    HIGHLIGHTBATCH = 300

    def run_tokenizer(self, batch=HIGHLIGHTBATCH):
        range_start, range_end = self._get_highlight_range()

        if self._highlight_done < range_start:
            self._highlight_done = range_start

        if not self._highlight_iter:
            f = max(range_start, self._highlight_done - 1)
            self._highlight_iter = syntax_highlight.begin_tokenizer(
                self.document, self.tokenizer, f)

        updatefrom = self.document.endpos()
        updateto = range_start
        updated = False
        finished = False

        for n, (f, t, style) in enumerate(self._highlight_iter):
            f = max(range_start, f)
            t = min(range_end, t)
            if f < t:
                self.document.styles.setints(f, t, style)

            updated = True
            updatefrom = min(f, updatefrom)
            updateto = max(t, updateto)

            if batch and (n > batch):
                break

            if t >= range_end:
                finished = True
                break
        else:
            finished = True

        if self.document.endpos() == 0 or updated and (updatefrom != updateto):
            self.document.style_updated(updatefrom, updateto)
            self._highlight_done = updateto

        # returns False if finished to terminate idle loop.
        return not finished

    def _split_chars(self, begin, end):
        """split characters by character category."""

        s = self.document.gettext(begin, end)

        def get_category(c):
            return unicodedata.category(c)[0]

        for key, chars in itertools.groupby(s, get_category):
            chars = ''.join(chars)
            end = begin + len(chars)
            yield begin, end, chars, key
            begin = end

    RE_WORDCHAR = r"(?P<L_WORDCHAR>[a-zA-Z0-9_]+)"
    RE_WHITESPACE = r"(?P<Z_WHITESPACE>[\t ]+)"
    RE_LF = r"(?P<_LF>\n)"
    RE_HIRAGANA = r"(?P<L_HIRAGANA>[\u3040-\u309f\u30fc]+)"
    RE_KATAKANA = r"(?P<L_KATAKANA>[\u30a0-\u30ff\u30fc]+)"

    RE_SPLITWORD = doc_re.compile('|'.join((
        RE_WORDCHAR, RE_WHITESPACE, RE_LF, RE_HIRAGANA, RE_KATAKANA)))

# http://unicodebook.readthedocs.org/en/latest/unicode.html
# Unicode 6.0 has 7 character categories, and each category has subcategories:
#
# Letter (L): lowercase (Ll), modifier (Lm), titlecase (Lt), uppercase (Lu), other (Lo)
# Mark (M): spacing combining (Mc), enclosing (Me), non-spacing (Mn)
# Number (N): decimal digit (Nd), letter (Nl), other (No)
# Punctuation (P): connector (Pc), dash (Pd), initial quote (Pi), final quote (Pf), open (Ps), close (Pe), other (Po)
# Symbol (S): currency (Sc), modifier (Sk), math (Sm), other (So)
# Separator (Z): line (Zl), paragraph (Zp), space (Zs)
# Other (C): control (Cc), format (Cf), not assigned (Cn), private use
# (Co), surrogate (Cs)

    def split_word(self, begin, all=False):
        """yield word in the document until line ends"""

        for m in self.RE_SPLITWORD.finditer(self.document, pos=begin):
            # split unmatched characters by character category.
            f, t = m.span()
            if f != begin:
                yield from self._split_chars(begin, f)
            begin = t

            yield (f, t, m.group(), m.lastgroup)

            # finish if we reach '\n'
            if not all and m.lastgroup == '_LF':
                return

        yield from self._split_chars(begin, self.document.endpos())

    def get_word_at(self, pos):
        tol = self.document.gettol(pos)
        prev = None
        for f, t, s, cg in self.split_word(tol):
            if pos == t:
                if cg not in {'Z_WHITESPACE', '_LF'}:
                    prev = (f, t, cg)
            elif pos < t:
                if cg not in {'Z_WHITESPACE', '_LF'}:
                    return (f, t, cg)
                else:
                    return prev

        return prev

    def get_word_list(self):
        words = collections.OrderedDict()
        for f, t, s, cg in self.split_word(0, all=True):
            if cg in {'Z_WHITESPACE', '_LF'}:
                continue
            if cg[0] in 'LMN':  # Letter, Mark, Number
                if len(s) > 2:
                    words[s] = None
        return list(words.keys())

    def search_next(self, pos, searchinfo):
        regex = searchinfo.get_regex()
        pos = min(max(0, pos), self.document.endpos())
        m = regex.search(self.document, pos)
        return m

    def search_prev(self, pos, searchinfo):
        regex = searchinfo.get_regex()
        last = None
        for m in regex.finditer(self.document, 0):
            span = m.span()
            if span[1] >= pos:
                break
            last = m
        return last

    def get_line_sel(self, wnd):
        doc = wnd.document
        sel = wnd.screen.selection.get_selrange()
        if not sel:
            f = t = wnd.cursor.pos
        else:
            f, t = sel

        tol = doc.gettol(f)

        t_tol = doc.gettol(t)
        if t != t_tol:
            eol = doc.geteol(t)
        else:
            eol = t

        return tol, eol

    def get_indent_range(self, pos):
        tol = self.document.gettol(pos)
        regex = doc_re.compile(self.RE_WHITESPACE)
        m = regex.match(self.document, tol)
        if m:
            return m.span()
        else:
            return (tol, tol)

    def build_indent_str(self, col):
        if self.indent_tab:
            ctab = col // self.tab_width
            cspc = col % self.tab_width
            return '\t' * ctab + ' ' * cspc
        else:
            return ' ' * col

    def get_parent_indent(self, pos):
        tol = self.document.gettol(pos)
        f, t = self.get_indent_range(tol)
        cols = self.calc_cols(f, t)

        while tol:
            eol = tol
            tol = self.document.gettol(eol - 1)
            f, t = self.get_indent_range(tol)
            pcols = self.calc_cols(f, t)
            if pcols < cols:
                return pcols

    def cancel_auto_indent(self, wnd):
        if self._last_autoindent:
            f, t = (min(max(0, p), self.document.endpos())
                    for p in self._last_autoindent)
            self._last_autoindent = None
            if f != t and wnd.cursor.pos == t:
                e = self.document.get_line_break(wnd.cursor.pos)
                if e == t:
                    s = self.document.gettext(f, t)
                    m = re.match(self.RE_WHITESPACE, s)
                    if m and m.group() == s:
                        self.delete_string(
                            wnd, f, t, update_cursor=False)
                        wnd.cursor.setpos(f)
                        return True

    def calc_next_indent(self, pos):
        return None

    MAX_AUTO_INDENT = 30

    def on_auto_indent(self, wnd):
        pos = wnd.cursor.pos
        f, t = self.get_indent_range(pos)
        if pos <= t:
            b = self.document.get_line_break(pos)
            self.insert_string(wnd, f, '\n', update_cursor=False)
            wnd.cursor.setpos(pos + 1)
            wnd.cursor.savecol()

            # if new line don't hava non-ws char,
            # thie line should be subject to cancel-indent.
            if t == b:
                self._last_autoindent = (f + 1, wnd.cursor.pos)
            return

        indentcol = self.calc_next_indent(pos)
        if indentcol is None:
            indent = self.document.gettext(f, min(pos, t))
        else:
            indent = self.build_indent_str(
                min(self.MAX_AUTO_INDENT, indentcol))

        self.insert_string(wnd, pos, '\n' + indent, update_cursor=False)
        self.delete_ws(wnd, pos + len(indent) + 1)
        wnd.cursor.setpos(pos + len(indent) + 1)
        wnd.cursor.savecol()
        self._last_autoindent = (pos + 1, wnd.cursor.pos)

    def calc_cols(self, f, t):
        chars = self.document.gettext(f, t)
        (dispchrs, dispcols, positions,
         intervals) = screen.translate_chars(f, chars, self.tab_width,
                                             kaa.app.config.AMBIGUOUS_WIDTH)

        return sum(dispcols)

    def on_global_next(self, wnd):
        return

    def on_global_prev(self, wnd):
        return



import curses

import kaa
import kaa.context
from kaa.cui.wnd import Window
from kaa.cui import splitter, editor


class ChildFrame(Window, kaa.context.ContextRoot):
    splitter = None
    active_editor = None

    def __init__(self, parent, pos=None):
        super().__init__(parent, pos=pos)

        self.set_label('frame', self)
        self.splitter = None

    def destroy(self):
        if self.splitter:
            self.splitter.destroy()

        super().destroy()

        self.destroy_context()
        self.mainframe.childframes.remove(self)

    def show_doc(self, doc):
        if self.splitter:
            self.splitter.destroy()
            self.splitter = None

        editorwnd = editor.TextEditorWindow(parent=self)
        editorwnd.show_doc(doc)
        self.splitter = splitter.Splitter(self, None, editorwnd)
        self.splitter.set_rect(*self.rect)
        return editorwnd

    def set_active_editor(self, editorwnd):
        self.active_editor = editorwnd

    def get_documents(self):
        docs = []
        for splitter, w in self.splitter.walk():
            if w:
                doc = w.document
                if doc not in docs:
                    docs.append(doc)
        return docs

    def get_title(self):
        if self.splitter:
            return '/'.join(self.splitter.get_title())
        return '<untitled>'

    def get_editors(self):
        if self.splitter:
            return self.splitter.get_editors()
        else:
            return []

    def bring_top(self):
        super().bring_top()
        if self.splitter:
            self.splitter.bring_top()
        self.mainframe.set_activeframe(self)
        self.mainframe.on_console_resized()

    def on_setrect(self, l, t, r, b):
        if self.splitter:
            self.splitter.set_rect(l, t, r, b)

    def on_focus(self):
        super().on_focus()
        self.mainframe.set_activeframe(self)

        if self.active_editor:
            self.active_editor.activate()
        elif self.splitter:
            self.splitter.activate()
        else:
            kaa.app.set_focus(self)

    def draw_screen(self):
        if kaa.app.get_activeframe() is not self:
            return

        if self.splitter:
            self.splitter.draw()


class MainFrame(Window, kaa.context.ContextRoot):

    """The main frame of application

    Attributes:
        editorwnd -- current editor wnd
    """

    popups = []
    inputline = None

    need_refresh = False
#    need_redraw = False
    childframes = []
    activeframe = None

    MESSAGEBAR_HEIGHT = 1
    BORDER_WIDTH = 1

    def __init__(self, wnd):
        Window.mainframe = self
        super().__init__(None, wnd)
        self.width, self.height = 0, 0

        self.messagebar = editor.TextEditorWindow(parent=self)

    def destroy(self):
        super().destroy()
        self.childframes = None
        self.activeframe = None
        self.destroy_context()

    def set_messagebar(self, doc):
        self.messagebar.show_doc(doc)

    def on_focus(self):
        super().on_focus()
        if self.activeframe:
            self.activeframe.on_focus()

    def on_console_resized(self):
        """Resize wnds"""
        self.height, self.width = self._cwnd.getmaxyx()
        self.messagebar.set_rect(
            0, self.height - self.MESSAGEBAR_HEIGHT, self.width, self.height)

        if self.inputline and not self.inputline.closed:
            w, h = self.inputline.getsize()
            editorheight = max(1, self.height - h - self.MESSAGEBAR_HEIGHT)
        else:
            editorheight = max(1, self.height - self.MESSAGEBAR_HEIGHT)

        if self.activeframe:
            self.activeframe.set_rect(0, 0, self.width, editorheight)

        for popup in self.popups:
            popup.on_console_resized()

    def _get_temporary_frame(self):
        # if kaa has only one frame and the frame has temporary document,
        # then dispose the document and reuse the frame.

        frames = kaa.app.get_frames()
        if len(frames) != 1:
            return

        wnd = frames[0].splitter.wnd
        if wnd and wnd.document:
            doc = wnd.document
            if doc.temporary and not wnd.document.undo.is_dirty():
                if len(doc.wnds) == 1:
                    return frames[0]

    def register_childframe(self, childframe):
        if childframe in self.childframes:
            self.childframes.remove(childframe)
        self.childframes.insert(0, childframe)

    def is_idle(self):
        return not self.popups and not self.inputline

    def show_doc(self, doc):
        frame = self._get_temporary_frame()
        if frame:
            if frame.splitter.wnd.document is doc:
                return
        else:
            frame = ChildFrame(parent=self)

        editorwnd = frame.show_doc(doc)
        editorwnd.activate()

        self.register_childframe(frame)

        self.set_activeframe(frame)
        self.on_console_resized()

        return editorwnd

    def set_activeframe(self, frame):
        self.activeframe = frame

    def add_popup(self, popup):
        self.popups.append(popup)

    def show_inputline(self, inputline):
        self.inputline = inputline
        self.show_dialog(inputline)
        self.on_console_resized()

    def show_dialog(self, dialog):
        self.add_popup(dialog)
        dialog.on_console_resized()
        dialog.start()

    def request_refresh(self):
        self.need_refresh = True

    def run_refresh(self):
        if self.need_refresh:
            #            if self.need_redraw:
            #                for child in self.walk_children():
            #                    child._cwnd.touchwin()

            curses.panel.update_panels()
            curses.doupdate()
            self.need_refresh = False
#            self.need_refresh = self.need_redraw = False

            return True

    def on_idle(self):
        if not self.closed:
            if self.run_refresh():
                return True

            for child in self.walk_children():
                if not child.closed:
                    if child.on_idle():
                        return True

            for child in self.walk_children():
                if not child.closed:
                    child.update_window()



from kaa.filetype.default import defaultmode
from kaa.theme import Style
from kaa.syntax_highlight import *

DiffThemes = {
    'basic': [
        Style('default', 'Base00', None),
        Style('fromfile', 'Orange', None),
        Style('tofile', 'Yellow', None),
        Style('hunk', 'Blue', None),
        Style('remove', 'Orange', None, bold=True),
        Style('add', 'Yellow', None, bold=True),
        Style('header', 'Orange', None),
    ],
}


def make_tokenizer():
    return Tokenizer(tokens=[
        ('fromfile', Span('fromfile', r'^---', '$')),
        ('tofile', Span('tofile', r'^\+\+\+', '$')),
        ('hunk', Span('hunk', r'^@@', '@@')),
        ('add', Span('add', r'^\+', '$')),
        ('remove', Span('remove', r'^-', '$')),
        ('header', Span('header', r'^\S+', '$')),
    ])


class DiffMode(defaultmode.DefaultMode):
    MODENAME = 'Diff'
    tokenizer = make_tokenizer()

    def init_themes(self):
        super().init_themes()
        self.themes.append(DiffThemes)



import kaa
from kaa import command
from kaa.ui.multiline import multilinemode
from kaa.keyboard import *


class PasteLinesMode(multilinemode.MultilineMode):

    auto_indent = False

    def close(self):
        super().close()
        self.target = None

    def on_add_window(self, wnd):
        super().on_add_window(wnd)

        kaa.app.messagebar.set_message("Paste text and hit alt+Enter")

    @command.commandid('multiline.done')
    @command.norec
    @command.norerun
    def paste_lines(self, wnd):
        f, t = self.document.marks['multiline']
        s = self.document.gettext(f, t)

        self.target.document.mode.put_string(self.target, s)

        kaa.app.messagebar.set_message("{} letters inserted".format(t - f))

        self.multiline_done(wnd)

    @classmethod
    def build(cls, target):
        doc = super().build()
        doc.mode.target = target

        return doc



import kaa_testutils
from kaa.filetype.markdown.markdownmode import MarkdownMode


class TestMarkdownHighlight(kaa_testutils._TestDocBase):
    DEFAULTMODECLASS = MarkdownMode
    TOKENIZER = MarkdownMode.tokenizer

    def test_header1(self):
        doc = self._getdoc('abc\n---')
        doc.mode.run_tokenizer(None)

        kaa_testutils.check_style(doc, 0, 7, 
            [self.TOKENIZER.tokens.header1] * 7)

    def test_header2(self):
        doc = self._getdoc('# abc')
        doc.mode.run_tokenizer(None)

        kaa_testutils.check_style(doc, 0, 5, 
            [self.TOKENIZER.tokens.header2] * 5)

    def test_hr(self):
        doc = self._getdoc('----')
        doc.mode.run_tokenizer(None)

        kaa_testutils.check_style(doc, 0, 4, 
            [self.TOKENIZER.tokens.hr] * 4)

    def test_link(self):
        doc = self._getdoc('[link][link]')
        doc.mode.run_tokenizer(None)
        kaa_testutils.check_style(doc, 0, 12, 
            [self.TOKENIZER.tokens.link]*12)

        doc = self._getdoc('[link](url)')
        doc.mode.run_tokenizer(None)
        kaa_testutils.check_style(doc, 0, 11, 
            [self.TOKENIZER.tokens.link]*7 + 
            [self.TOKENIZER._LinkTokenizer.tokens.default]*3+
            [self.TOKENIZER._LinkTokenizer.tokens.close]*1)

        doc = self._getdoc('[link](url "te)xt")')
        doc.mode.run_tokenizer(None)
        kaa_testutils.check_style(doc, 0, 19, 
            [self.TOKENIZER.tokens.link]*7 + 
            [self.TOKENIZER._LinkTokenizer.tokens.default]*4+
            [self.TOKENIZER._LinkTokenizer.tokens.desc]*7+
            [self.TOKENIZER._LinkTokenizer.tokens.close]*1)

        doc = self._getdoc('[link]:')
        doc.mode.run_tokenizer(None)
        kaa_testutils.check_style(doc, 0, 7, 
            [self.TOKENIZER.tokens.link]*7)

    def test_image(self):
        doc = self._getdoc('![link][link]')
        doc.mode.run_tokenizer(None)
        kaa_testutils.check_style(doc, 0, 13,
            [self.TOKENIZER.tokens.link]*13)

        doc = self._getdoc('![link](url)')
        doc.mode.run_tokenizer(None)
        kaa_testutils.check_style(doc, 0, 12,
            [self.TOKENIZER.tokens.link]*8 + 
            [self.TOKENIZER._LinkTokenizer.tokens.default]*3 + 
            [self.TOKENIZER._LinkTokenizer.tokens.close]*1)

    def test_emphasis(self):
        doc = self._getdoc('**text**')
        doc.mode.run_tokenizer(None)
        kaa_testutils.check_style(doc, 0, 8,
            [self.TOKENIZER.tokens.strong1]*8)

        doc = self._getdoc('__text__')
        doc.mode.run_tokenizer(None)
        kaa_testutils.check_style(doc, 0, 8,
            [self.TOKENIZER.tokens.strong2]*8)

        doc = self._getdoc('*text*')
        doc.mode.run_tokenizer(None)
        kaa_testutils.check_style(doc, 0, 6,
            [self.TOKENIZER.tokens.emphasis1]*6)

        doc = self._getdoc('_text_')
        doc.mode.run_tokenizer(None)
        kaa_testutils.check_style(doc, 0, 6,
            [self.TOKENIZER.tokens.emphasis2]*6)

    def test_literal(self):
        doc = self._getdoc('`text`')
        doc.mode.run_tokenizer(None)
        kaa_testutils.check_style(doc, 0, 6,
            [self.TOKENIZER.tokens.code3]*6)

        doc = self._getdoc('```\ntext\n```\n')
        doc.mode.run_tokenizer(None)
        kaa_testutils.check_style(doc, 0, 13,
            [self.TOKENIZER.tokens.code2]*13)

        doc = self._getdoc('` text`')
        doc.mode.run_tokenizer(None)
        kaa_testutils.check_style(doc, 0, 7,
            [self.TOKENIZER.tokens.default]*6 +
            [self.TOKENIZER.tokens.code3])

        doc = self._getdoc('    text\na')
        doc.mode.run_tokenizer(None)
        kaa_testutils.check_style(doc, 0, 10,
            [self.TOKENIZER.tokens.code1]*8 +
            [self.TOKENIZER.tokens.default]*2)

    def test_list(self):
        doc = self._getdoc('* abc')
        doc.mode.run_tokenizer(None)
        kaa_testutils.check_style(doc, 0, 5,
            [self.TOKENIZER.tokens.list]*2 +
            [self.TOKENIZER.tokens.default]*3)

        doc = self._getdoc('1. abc')
        doc.mode.run_tokenizer(None)
        kaa_testutils.check_style(doc, 0, 6,
            [self.TOKENIZER.tokens.list]*3 +
            [self.TOKENIZER.tokens.default]*3)

        doc = self._getdoc('1. abc\n    * def')
        doc.mode.run_tokenizer(None)
        kaa_testutils.check_style(doc, 0, 16,
            [self.TOKENIZER.tokens.list]*3 +
            [self.TOKENIZER.tokens.default]*4+
            [self.TOKENIZER.tokens.list]*6+
            [self.TOKENIZER.tokens.default]*3)

        doc = self._getdoc('1. abc\na\n    * def')
        doc.mode.run_tokenizer(None)
        kaa_testutils.check_style(doc, 0, 18,
            [self.TOKENIZER.tokens.list]*3 +
            [self.TOKENIZER.tokens.default]*6+
            [self.TOKENIZER.tokens.code1]*9)




import re
from kaa import cursor, keyboard
from kaa.filetype.default import modebase
from kaa.theme import Style
from kaa.exceptions import KaaError

DialogThemes = {
    'basic': [
        Style('default', 'Base3', 'Base02', False, False),
        Style('caption', 'Orange', 'Base02', nowrap=True),
        Style('blank-line-header', 'Blue', None),

        Style('activemark', 'Base02', 'Yellow', nowrap=True),
        Style('nonactivemark', 'Yellow', 'Base02', nowrap=True),

        Style('button', 'Base3', 'Base01', nowrap=True),
        Style('button.shortcut', 'Base3', 'Base01', underline=True,
              nowrap=True),

        # todo: right-button with rjust should be defined in another name.
        Style('right-button', 'Base3', 'Base01', rjust=True, nowrap=True),
        Style('right-button.checked', 'Base3', 'Orange', rjust=True,
              nowrap=True),
        Style('right-button.shortcut', 'Base3', 'Base01', underline=True,
              rjust=True, nowrap=True),
        Style('right-button.shortcut.checked', 'Base3', 'Orange',
              underline=True, rjust=True, nowrap=True),

        Style('selectitem', 'Cyan', 'Base02', nowrap=True),
        Style('selectitem2', 'Yellow', 'Base02', nowrap=True),
        Style('selectitem-active', 'Base02', 'Yellow', nowrap=True),

        Style('selectphrase', 'Black', 'Cyan', nowrap=True),
        Style('selectphrase2', 'Yellow', 'Base02', nowrap=True),
        Style('selectphrase-active', 'Base02', 'Yellow', nowrap=True),

        Style('dialog-border', 'Blue', 'Base03'),
    ],
}


class DialogCursor(cursor.Cursor):

    def __init__(self, wnd, ranges):
        super().__init__(wnd)
        self.ranges = ranges

    def _findrange(self, curpos):
        for func in self.ranges:
            f, t = func(self)
            if f <= curpos <= t:
                return f, t
        return self.ranges[0](self)

    def adjust_nextpos(self, curpos, nextpos):
        f, t = self._findrange(curpos)
        if nextpos >= t:
            return t
        if nextpos <= f:
            return f
        return nextpos


def _MarkRange(mark, cursor):
    return cursor.wnd.document.marks[mark]


def MarkRange(mark):
    return lambda cursor: _MarkRange(mark, cursor)


class DialogMode(modebase.ModeBase):
    NO_WRAPINDENT = False
    DEFAULT_STATUS_MSG = ''

    autoshrink = False
    suspend_autoshrink = False
    stack_upper = True
    border = False
    min_height = 1

    def init_themes(self):
        super().init_themes()
        self.themes.append(DialogThemes)

    def init_tokenizers(self):
        self.tokenizers = []

    def on_add_window(self, wnd):
        super().on_add_window(wnd)

    def calc_width(self, wnd):
        ret = wnd.mainframe.getsize()[0]
        if self.border:
            ret -= wnd.mainframe.BORDER_WIDTH * 2
        return ret

    def calc_height(self, wnd):
        maxw, maxh = wnd.mainframe.getsize()
        height = wnd.screen.get_total_height(maxh)
        ret = min(maxh, height)
        if height < self.min_height:
            return self.min_height
        return ret

    def calc_position(self, wnd):
        w = self.calc_width(wnd)

        # temporary set screen size
        # should be resized later soon
        wnd.screen.setsize(w, wnd.screen.height)

        height = self.calc_height(wnd)

        if not self.border:
            height = min(height, wnd.mainframe.messagebar.rect[1])
            top = wnd.mainframe.messagebar.rect[1] - height
            return 0, top, wnd.mainframe.width, top + height
        else:
            border_width = wnd.mainframe.BORDER_WIDTH
            height = min(
                height,
                wnd.mainframe.messagebar.rect[1] - border_width * 2)
            top = wnd.mainframe.messagebar.rect[1] - height - border_width
            return (
                border_width, top, wnd.mainframe.width - border_width, top + height
            )

    def resize_inputs(self, wnd, inputs):
        rects = [w.document.mode.calc_position(w) for w in inputs]
        rc_input = self.calc_position(wnd)

        heights = [b - t for (l, t, r, b) in rects]

        l, t, r, b = rc_input

        if self.stack_upper:
            t = t - sum(heights)
        else:
            t = b
        if t < 0:
            raise KaaError(
                'Invalid window position {!r}:{!r}/{!r}'.format(self, rects, rc_input))

        positions = []
        for wl, wt, wr, wb in rects:
            b = t + (wb - wt)
            positions.append((l, t, r, b))
            t = b + 1

        return rc_input, positions

    def on_dialog_start(self, wnd):
        pass

    def run_autoshrink(self):
        for wnd in self.document.wnds:
            w, h = wnd.getsize()
            l, t, r, b = self.calc_position(wnd)
            newh = b - t
            # resize window if number of rows changed
            if newh != h:
                wnd.get_label('popup').on_console_resized()
                wnd.mainframe.on_console_resized()
                f, t = wnd.screen.get_visible_range()
                if t == self.document.endpos():
                    # display as many rows as we can
                    wnd.screen.locate(t, bottom=True, align_always=True)
                    wnd.cursor.setpos(wnd.cursor.pos)

    def on_document_updated(self, pos, inslen, dellen):
        super().on_document_updated(pos, inslen, dellen)

        if self.autoshrink:
            if not self.suspend_autoshrink:
                self.run_autoshrink()


class FormBuilder:
    re_accel = re.compile(r'&.')

    def __init__(self, document):
        self.document = document
        self.doc_autoshrink = getattr(self.document.mode, 'autoshrink', None)

    def __enter__(self):
        if self.doc_autoshrink:
            self.document.mode.suspend_autoshrink = True
        return self

    def __exit__(self, type, value, traceback):
        if getattr(self.document.mode, 'suspend_autoshrink', False):
            self.document.mode.suspend_autoshrink = False
            self.document.mode.run_autoshrink()

    def append_text(
            self, stylename, text, mark_start=None, mark_end=None,
            mark_pair=None, on_shortcut=None, shortcut_style=None,
            shortcut_mark=None, shortcut_need_alt=True):

        # marks should not be moved until the document completed.
        self.document.marks.locked = True
        try:
            start = self.document.endpos()

            if shortcut_style is None:
                shortcut_style = stylename

            if stylename:
                style_id = self.document.mode.get_styleid(stylename)

            if shortcut_style:
                shortcut_style_id = self.document.mode.get_styleid(
                    shortcut_style)

            # convert & escape to char and build shortcut dict
            lastpos = 0
            for m in self.re_accel.finditer(text):
                span = m.span()
                if span[0] != lastpos:
                    f = self.document.endpos()
                    self.document.append(text[lastpos:span[0]])
                    if stylename:
                        self.document.setstyles(
                            f, self.document.endpos(), style_id, update=False)

                f = self.document.endpos()
                c = m.group()[1]
                self.document.append(c)
                if c != '&':
                    if on_shortcut:
                        self.document.mode.keybind.add_keybind(
                            {(keyboard.alt, c.lower()): on_shortcut})
                        if not shortcut_need_alt:
                            self.document.mode.keybind.add_keybind(
                                {(c.lower()): on_shortcut})
                            self.document.mode.keybind.add_keybind(
                                {(c.upper()): on_shortcut})

                    if shortcut_mark:
                        self.document.marks[shortcut_mark] = f

                    if shortcut_style:
                        self.document.setstyles(
                            f,
                            f + 1,
                            shortcut_style_id,
                            update=False)

                lastpos = m.end()

            else:
                if lastpos != len(text):
                    f = self.document.endpos()
                    self.document.append(text[lastpos:])
                    if stylename:
                        self.document.setstyles(f, self.document.endpos(),
                                                style_id, update=False)

            if mark_start is not None:
                self.document.marks[mark_start] = start

            if mark_end is not None:
                self.document.marks[mark_end] = self.document.endpos()

            if mark_pair is not None:
                self.document.marks[mark_pair] = (
                    start, self.document.endpos())

            # Dialog texts are not undoable
            if self.document.undo:
                self.document.undo.clear()

        finally:
            self.document.marks.locked = False



from unittest.mock import patch

import kaa
from kaa.ui.searchdlg import searchdlgmode
import kaa_testutils


class TestSearchDlgMode(kaa_testutils._TestDocBase):

    def _getmodeclass(self):
        return lambda: searchdlgmode.SearchDlgMode(
            target=kaa_testutils._TestScreenBase()._getwnd(''))

    def test_searchdlg(self):
        doc = self._getdoc('')
        doc.mode.option.text = ''

        doc.mode.build_document()
        ignorecase = doc.mode.option.ignorecase
        word = doc.mode.option.word
        regex = doc.mode.option.regex

        doc.mode.toggle_option_ignorecase(None)
        doc.mode.toggle_option_word(None)
        doc.mode.toggle_option_regex(None)

        assert ignorecase != doc.mode.option.ignorecase
        assert word != doc.mode.option.word
        assert regex != doc.mode.option.regex

        pos = doc.marks['searchtext']
        doc.insert(pos[0], 'test string')

        assert doc.mode.get_search_str() == 'test string'

    def test_execsearh(self):
        doc = self._getdoc('')
        doc.mode.build_document()

        option = doc.mode.option
        option.ignorecase = False
        option.word = False
        option.regex = False
        pos = doc.marks['searchtext']
        doc.insert(pos[0], 'test string')

        doc.mode.target.document.append('012345test string')
        doc.mode.search_next(None)
        assert doc.mode.target.screen.selection.get_selrange() == (6, 17)


class TestReplaceDlg(kaa_testutils._TestDocBase):

    def _getmodeclass(self):
        return lambda: searchdlgmode.ReplaceDlgMode(
            target=kaa_testutils._TestScreenBase()._getwnd(''))

    def test_replacedlg(self):

        doc = self._getdoc('')
        doc.mode.option.text = ''
        doc.mode.build_document()

        ignorecase = doc.mode.option.ignorecase
        word = doc.mode.option.word
        regex = doc.mode.option.regex

        doc.mode.toggle_option_ignorecase(None)
        doc.mode.toggle_option_word(None)
        doc.mode.toggle_option_regex(None)

        assert ignorecase != doc.mode.option.ignorecase
        assert word != doc.mode.option.word
        assert regex != doc.mode.option.regex

        pos = doc.marks['searchtext']
        doc.insert(pos[0], 'test string')

        assert doc.mode.get_search_str() == 'test string'

    def test_execsearh(self):
        kaa.app.DEFAULT_THEME = 'basic'

        doc = self._getdoc('')
        doc.mode.option.text = ''
        doc.mode.build_document()

        option = doc.mode.option
        option.ignorecase = False
        option.word = False
        option.regex = False
        pos = doc.marks['searchtext']
        doc.insert(pos[0], 'test string')

        doc.mode.target.document.append('012345test string')
        doc.mode.search_next(None)
        assert doc.mode.target.screen.selection.get_selrange() == (6, 17)



from kaa.filetype import filetypedef


class FileTypeInfo(filetypedef.FileTypeInfo):
    FILE_EXT = {'.h', '.c'}

    @classmethod
    def get_modetype(cls):
        from kaa.filetype.c.cmode import CMode
        return CMode



import types
from kaa import doc_re


def begin_tokenizer(doc, root, updated):
    updated = min(updated, doc.endpos() - 1)
    # find default token
    while updated >= 1:
        token = root.get_token_at(doc, updated)
        pos = token.get_token_begin(doc, updated)
        if not isinstance(token, DefaultToken):
            updated = pos - 1
            continue

        is_resumable = True
        for parent in token.get_tokenizers():
            if not parent.is_resumable:
                is_resumable = False
                break

        if not is_resumable:
            updated = pos - 1
            continue

        for parent in token.get_tokenizers():
            pos = yield from parent.run(doc, pos)

        return pos

    else:
        pos = yield from root.run(doc, 0)
        return pos


class Token:
    styles = ()

    def __init__(self, terminates=False):
        #        self._stylename = stylename
        self._style_ids = []
        self._styleid_names = {}
        self.terminates = terminates

    def register_styles(self, styles):
        l = self.tokenizer.get_styleid_map()
        for stylename, styleidname in styles:
            styleid = len(l)
            self._style_ids.append(styleid)
            l[styleid] = self
            self._styleid_names[styleid] = stylename
            setattr(self, styleidname, styleid)

    def get_tokenizers(self):
        tokenizer = self.tokenizer
        while tokenizer:
            yield tokenizer
            tokenizer = tokenizer.parent

    def get_stylename(self, styleid):
        return self._styleid_names[styleid]

    def get_token_begin(self, doc, pos):
        p = doc.styles.rfindint(self._style_ids, 0, pos,
                                comp_ne=True)
        if p == -1:
            return 0
        else:
            return p + 1

    def set_tokenizer(self, tokenizer):
        self.tokenizer = tokenizer
        self.prepare()


class DefaultToken(Token):

    def __init__(self, stylename, terminates=False):
        super().__init__(terminates=terminates)
        self._stylename = stylename

    def prepare(self):
        self.register_styles([(self._stylename, "styleid_default")])

    def re_start(self):
        return ''


class Tokenizer:
    DEFAULT_TOKEN = DefaultToken
    IS_RESUMABLE = True
    re_starts = None

    def __init__(self, *, parent=None, default_style='default', terminates=None,
                 is_resumable=True, tokens=()):

        self.parent = parent
        if not parent:
            self.styleid_map = {}  # this is root tokenizer

        self.terminates = terminates
        self.is_resumable = is_resumable
        self._token_list = []
        self.tokens = types.SimpleNamespace()

        self.tokens.default = self.DEFAULT_TOKEN(default_style)
        self.tokens.default.set_tokenizer(self)
        self.styleid_default = self.tokens.default.styleid_default

        for name, token in tokens:
            self.add_token(name, token)

        self.prepare()

    def add_token(self, name, token):
        self._token_list.append(token)
        setattr(self.tokens, name, token)
        token.set_tokenizer(self)

    def get_styleid_map(self):
        if self.parent:
            return self.parent.get_styleid_map()
        else:
            return self.styleid_map

    def prepare(self):
        self.groupnames = {}
        starts = []
        for i, token in enumerate(self._token_list):
            start = token.re_start()
            if start:
                name = 'G{}'.format(i)
                self.groupnames[name] = token
                starts.append(r'(?P<{}>{})'.format(name, start))

        if self.terminates:
            starts.insert(
                0, r'(?P<TERMINATE>{})'.format(self.terminates))
            self.groupnames['TERMINATE'] = None

        if starts:
            self.re_starts = doc_re.compile(
                '|'.join(starts),
                doc_re.M + doc_re.X)

    def resume(self, doc, updated):
        pos = yield from self.run(doc, updated)
        if self.parent:
            pos = yield from self.parent.run(doc, pos)
        return pos

    def run(self, doc, pos):
        if not self.re_starts:
            return doc.endpos()

        while True:
            m = self.re_starts.search(doc, pos)
            if not m:
                break

            f, t = m.span()
            if f != pos:
                yield (pos, f, self.styleid_default)
                pos = f

            if m.lastgroup == 'TERMINATE':
                return f

            child = self.groupnames[m.lastgroup]
            pos, terminates = yield from child.on_start(doc, m)
            if terminates:
                return pos

        if pos != doc.endpos():
            yield (pos, doc.endpos(), self.styleid_default)
            pos = doc.endpos()

        return pos

    def get_token_at(self, doc, pos):
        styleid = doc.styles.getints(pos, pos + 1)[0]
        return self.get_styleid_token(styleid)

    def get_styleid_token(self, styleid):
        return self.get_styleid_map()[styleid]


# class Root(Tokenizer):
#
#    def __init__(self, tokens=(), default_style='default'):
#        self.styleid_map = {}
#        super().__init__(None, tokens=tokens, default_style=default_style, terminates=None)


class SingleToken(Token):

    def __init__(self, stylename, tokens, terminates=False):
        super().__init__(terminates=terminates)
        self._stylename = stylename
        self._tokens = tokens

    def prepare(self):
        self.register_styles([(self._stylename, "styleid_token")])

    def re_start(self):
        return r'({})'.format('|'.join(self._tokens))

    def on_start(self, doc, match):
        yield (match.start(), match.end(), self.styleid_token)
        return match.end(), self.terminates


class Keywords(SingleToken):

    def re_start(self):
        return (
            r'\b({})\b'.format('|'.join(doc_re.escape(k)
                                        for k in self._tokens))
        )


class Span(Token):

    def __init__(self, stylename, start, end, escape=None,
                 capture_end=True, terminates=None, terminate_tokens=None):

        self._stylename = stylename
        self._start = start
        self._escape = escape
        self._end = end
        self._capture_end = capture_end
        self._terminates = terminates
        self._terminate_tokens = terminate_tokens

        super().__init__(terminates=terminates)

    def prepare(self):
        self.register_styles([(self._stylename, "styleid_span")])

        end = self._end
        if self.tokenizer.terminates:
            end = '(?P<TERMINATES>{})|({})'.format(
                self.tokenizer.terminates, end)

        if self._terminate_tokens:
            end = '(?P<TERMINATES2>{})|({})'.format(
                self._terminate_tokens, end)

        if self._escape:
            end = '(?P<ESCAPE>{}.)|({})'.format(
                doc_re.escape(self._escape), end)
        self._re_end = doc_re.compile(end, doc_re.X + doc_re.M + doc_re.S)

    def re_start(self):
        return self._start

    def _is_span_end(self, doc, m):
        return True

    def on_start(self, doc, match):
        pos = match.end()
        yield (match.start(), pos, self.styleid_span)

        for m in self._re_end.finditer(doc, pos):
            if self._escape and m.group('ESCAPE') is not None:
                continue

            if not self._is_span_end(doc, m):
                continue

            if pos != m.start():
                yield (pos, m.start(), self.styleid_span)

            if self.tokenizer.terminates and m.group('TERMINATES') is not None:
                return m.start(), True

            if self._terminate_tokens and m.group('TERMINATES2') is not None:
                return m.start(), True

            if self._capture_end:
                if m.start() != m.end():
                    yield (m.start(), m.end(), self.styleid_span)
                return m.end(), self.terminates
            else:
                return m.start(), self.terminates
        else:
            if pos != doc.endpos():
                yield (pos, doc.endpos(), self.styleid_span)
            return doc.endpos(), self.terminates


class Terminator(SingleToken):

    def on_start(self, doc, match):
        if 0:  # make this fuction iterator
            yield ()
        return match.start(), True

# class HTMLTag(Token):
#    def on_start(self, doc, m, endpos):
#        angle_f, angle_t = m.span(1)
#        yield (angle_f, angle_t, self.styles.angle)
#
#        name_f, name_t = m.span(2)
#        if angle_t != name_f:
#            yield (angle_t, name_f, self.styles.null)
#
#        yield name_f, name_t, self.styles.elementname
#        pos = yield from self.run(doc, name_t, endpos, ns)
#
#        elemname = m.group(2).lower()
#        if elemname == 'script':
#            type = ns.attrs.get('type', '').lower().strip()
#            if not type or 'text/javascript' == type:
#                pos = yield from self.javascript.run(doc, pos, endpos)
#
#        elif elemname == 'style':
#            type = ns.attrs.get('type', '').lower().strip()
#            if not type or 'text/javascript' == type:
#                pos = yield from self.css.run(doc, pos, endpos)
#
#        return pos
#
#



import curses
import curses.panel
import collections

from kaa import log
from kaa.cui import keydef
import kaa.context


class Window(kaa.context.Context):

    """Base class to wrapper of curses.window object

    Class attributes:
        mainframe -- Mainframe of the window
    """

    mainframe = None

    WAITINPUTFOR = 1000  # wait 1000ms for input
    closed = False
    editmode = None
    cursor = None

    def __init__(self, parent, wnd=None, pos=None):
        """Wrap window object wnd. Create new window if wnd was omitted."""

        super().__init__()

        w, h = 1, 1
        x = y = 0
        if parent:
            w, h = parent.getsize()

        if wnd:
            self._cwnd = wnd
        else:
            if pos:
                x, y = pos
            else:
                x = y = 0

            w = max(1, w)
            h = max(1, h)
            self._cwnd = curses.newwin(h, w, y, x)

        self._panel = curses.panel.new_panel(self._cwnd)
        self._cwnd.keypad(1)
        self.parent = parent
        self.children = []
        if self.parent:
            self.parent.children.append(self)

        self.rect = (x, y, x + w, y + h)

        self._oninit()

    def _oninit(self):
        pass

    def walk_children(self):
        # breadth first
        stack = collections.deque([self])
        while stack:
            v = stack.pop()
            if v is not self:
                yield v
            stack.extendleft(v.children)

    def destroy(self):
        """Destroy window"""
        if self.closed:
            return
        self.closed = True
        while self.children:
            self.children[0].destroy()

        if self.parent:
            self.parent.children.remove(self)
        self.parent = None
        self._cwnd = self._panel = None

        if kaa.app.focus is self:
            kaa.app.focus = None

    def get_context_parent(self):
        return self.parent

    _lasterror = ''
    _skipped = 0

    def restore_cursor_pos(self):
        if self.cursor:
            x, y = self.cursor.last_screenpos
            if (y, x) != self._cwnd.getyx():
                h, w = self._cwnd.getmaxyx()
                if y < h and x < w and y >= 0 and x >= 0:
                    self._cwnd.move(y, x)

    _keys = []
    _input_lasterror = ''
    _input_skipped = 0

    @classmethod
    def read_console(cls, cwnd, nonblocking):
        if nonblocking:
            cwnd.timeout(0)
        else:
            cwnd.timeout(-1)

        curses.panel.update_panels()
        curses.doupdate()

        last = None
        try:
            while True:
                c = cwnd.get_wch()

                cls._input_lasterror = ''
                cls._input_skipped = 0

                if c == curses.KEY_MOUSE:
                    mouseid, x, y, z, bstate = curses.getmouse()
                    cls._keys.append(keydef.MouseEvent(
                        mouseid, x, y, z, bstate))
                else:
                    cls._keys.extend(keydef.KeyEvent(k, True)
                                     for k in keydef.convert_registered_key(c))
                    last = cls._keys[-1]

                cwnd.timeout(0)

        except curses.error as e:
            if repr(e.args) != cls._input_lasterror:
                if e.args[0] != 'no input':
                    log.debug('Error in get_wch()', exc_info=True)
                cls._input_lasterror = repr(e.args)
            else:
                cls._input_skipped += 1
                if cls._input_skipped % 500 == 0:
                    log.debug(
                        'Too much input-error skips!: {} times'.format(cls._input_skipped))

        # last key event has no trailing character input(i.e. this is not key
        # sequqnce like function keys).
        if last:
            last.has_trailing_char = False

    def do_input(self, nonblocking):
        """Get an input from curses"""

        self._panel.top()

        self.read_console(self._cwnd, nonblocking)

        ret = []
        if self._keys:
            ret.append(self._keys.pop(0))
        return ret

    def add_str(self, letters, attr):
        try:
            self._cwnd.addstr(letters, attr)
        except curses.error:
            # exception is raised if a characters are written
            # over right border.
            #            logstr = letters if (len(letters) < 10) else (letters[:10]+'...')
            #            log.debug('Error to write str:{!r}'.format(logstr), exc_info=True)
            pass

    def bring_top(self):
        self._panel.top()

    def activate(self):
        """Activate wnd"""
        self.bring_top()
        kaa.app.set_focus(self)

    def set_rect(self, l, t, r, b):
        """Set window coordinate"""

        if hasattr(self, 'splitter') and self.splitter:
            assert t <= self.splitter.rect[1]

        try:
            self._cwnd = curses.newwin(b - t, r - l, t, l)
            self.rect = (l, t, r, b)
        except curses.error:
            log.debug('error on resizing window: {} {}'.format(self,
                                                               (l, t, r, b), exc_info=True))
            # shrink window to avoid segfault in curses
            self._cwnd = curses.newwin(1, 1, 0, 0)
            self.rect = (0, 0, 1, 1)

        # surprisingly enough, the following assert fails sometime.
#        t, l = self._cwnd.getbegyx()
#        h, w = self._cwnd.getmaxyx()
#        assert (l, t, l+w, t+h) == self.rect

        self._cwnd.keypad(1)

        self._panel = curses.panel.new_panel(self._cwnd)
        self.on_setrect(*self.rect)

    def draw_screen(self, force=False):
        pass

    def getsize(self):
        h, w = self._cwnd.getmaxyx()
        return w, h

    def refresh(self):
        self.mainframe.request_refresh()

    def update_window(self):
        return self.draw_screen()

    def update_status(self):
        return False

    def on_idle(self):
        return

    def on_setrect(self, l, t, r, b):
        """Window resized"""

    def on_focus(self):
        """Got focus"""

    def on_killfocus(self):
        pass



import kaa
import string


class EditMode:
    MODENAME = 'Insert'

    pending_str = ''
    repeat = None
    repeat_str = ''

    _replace_str = False
    _key_hook = None
    _post_key_hook = None

    def __init__(self):
        self.pending_keys = []
        self.last_command_keys = []

    def install_keyhook(self, hook):
        self._key_hook = hook

    def install_post_key_hook(self, hook):
        self._post_key_hook = hook

    def activated(self, wnd):
        pass

    def flush_pending_str(self, wnd):
        if self.pending_str:
            pending = self.pending_str
            self.pending_str = ''
            wnd.document.mode.on_str(wnd, pending, self._replace_str)
            return True

    def on_keyevent(self, wnd, event):
        if self._key_hook:
            f = self._key_hook
            self._key_hook = None
            event = f(wnd, event)

        if not event:
            return

        if event.key == '\x1b' and not event.has_trailing_char:
            return self.on_esc_pressed(wnd, event)
        else:
            return self.on_key_pressed(wnd, event)

    def _get_keybinds(self, wnd):
        return [wnd.document.mode.keybind]

    def _get_command(self, wnd, event):
        self.pending_keys.append(event.key)
        candidates = [(keys, command)
                      for keybind in self._get_keybinds(wnd)
                      for keys, command
                      in keybind.get_candidates(self.pending_keys)]

        if not candidates:
            if len(self.pending_keys) == 1:
                if isinstance(event.key, str):
                    s = event.key
                    return s, None, candidates

        elif len(candidates[0][0]) == len(self.pending_keys):
            return None, candidates[0][1], candidates

        return None, None, candidates

    def _on_str(self, wnd, s):
        if not wnd.document.mode.DELAY_STR:
            self.flush_pending_str(wnd)

        for c in s:
            if c in '\t\r\n' or c >= ' ':
                self.pending_str += c

                if not wnd.document.mode.DELAY_STR:
                    self.flush_pending_str(wnd)

    def _keys_to_str(self, keys):
        ret = []
        alt = False
        for k in keys:
            if k == '\x1b':
                alt = True
            else:
                s = kaa.app.get_keyname(k)
                if alt:
                    s = 'alt-' + s
                    alt = False
                ret.append(s)
        return ret

    def _show_pending_keys(self, wnd, s, commands, candidates):
        if s or commands or not candidates or not self.pending_keys:
            msg = wnd.document.mode.DEFAULT_MENU_MESSAGE or kaa.app.DEFAULT_MENU_MESSAGE
            kaa.app.messagebar.set_message(msg)
        else:
            cur = self._keys_to_str(self.pending_keys)
            curlen = len(cur)

            nextkeys = set()
            for k, commands in candidates:
                keys = self._keys_to_str(k)
                if len(keys) > curlen:
                    nextkeys.add(keys[curlen:curlen + 1][0])

            msg = ' '.join(cur) + ' [%s]' % (', '.join(sorted(nextkeys)))
            kaa.app.messagebar.set_message(msg)

    def on_non_command_str(self, wnd, event, s):
        self._on_str(wnd, s)
        del self.last_command_keys[:]

    def on_key_pressed(self, wnd, event):
        s, commands, candidate = self._get_command(wnd, event)
        s, commands, candidate = wnd.document.mode.on_keypressed(
            wnd, event, s, commands, candidate)

        self._show_pending_keys(wnd, s, commands, candidate)
        try:
            if s:
                self.on_non_command_str(wnd, event, s)

            elif commands:
                self.flush_pending_str(wnd)

                self.last_command_keys.append(self.pending_keys)
                del self.last_command_keys[:-3]

                wnd.document.mode.on_commands(wnd, commands, self.get_repeat())
                self.clear_repeat()

        finally:
            if self._post_key_hook and (s or commands):
                f = self._post_key_hook
                self._post_key_hook = None

                # flush pending str before hook runs.
                self.flush_pending_str(wnd)
                event = f(wnd, s, commands)

            if s or commands or not candidate:
                self.pending_keys = []

    def on_esc_pressed(self, wnd, event):
        kaa.app.messagebar.set_message('')
        self.pending_keys = []
        self.clear_repeat()

        self._key_hook = None
        self._post_key_hook = None

        del self.last_command_keys[:]

        if not wnd.closed:
            self.flush_pending_str(wnd)

        if not wnd.closed:
            # cancel mark
            wnd.screen.selection.set_mark(None)

            wnd.document.mode.on_esc_pressed(wnd, event)

    def clear_repeat(self):
        self.repeat = None
        self.repeat_str = ''

    def add_repeat_char(self, wnd, c):
        self.repeat_str += c
        try:
            self.set_repeat(int(self.repeat_str))
        except ValueError:
            self.init_repeat()

    def has_repeat(self):
        return self.repeat is not None

    def get_repeat(self):
        return self.repeat or 1

    def set_repeat(self, n):
        self.repeat = n


class ReplaceMode(EditMode):
    MODENAME = 'Replace'
    _replace_str = True


class CommandMode(EditMode):
    MODENAME = 'Command'

    def on_non_command_str(self, wnd, event, s):
        is_available, command = wnd.document.mode.get_command(
            'editmode.insert')
        if command:
            self.flush_pending_str(wnd)
            command(wnd)
            wnd.document.mode.on_str(wnd, s, False)
        else:
            super().on_non_command_str(wnd, event, s)

    def on_key_pressed(self, wnd, event):
        # check if key is repeat count
        if not self.pending_keys:
            if (isinstance(event.key, str) and (event.key > '0') and
                    (event.key in string.digits)):
                self.add_repeat_char(wnd, event.key)
                return

            if (event.key == '0') and self.repeat_str:
                self.add_repeat_char(wnd, event.key)
                return

        super().on_key_pressed(wnd, event)

    def _get_keybinds(self, wnd):
        return [wnd.document.mode.keybind_vi_commandmode,
                wnd.document.mode.keybind]

    def flush_pending_str(self, wnd):
        self.pending_str = ''


class VisualMode(CommandMode):
    MODENAME = 'Visual'

    def _get_keybinds(self, wnd):
        return [wnd.document.mode.keybind_vi_visualmode]


class VisualLinewiseMode(CommandMode):
    MODENAME = 'Visual(Line)'

    def _get_keybinds(self, wnd):
        return [wnd.document.mode.keybind_vi_visuallinewisemode]



import kaa_testutils
from kaa.filetype.html import htmlmode
from kaa.filetype.html.htmlmode import get_encoding


class TestHTMLHighlight(kaa_testutils._TestDocBase):
    DEFAULTMODECLASS = htmlmode.HTMLMode
    TOKENIZER = htmlmode.HTMLMode.tokenizer
    AttrTokenizer = TOKENIZER.AttrTokenizer

    ValueTokenizer1 = TOKENIZER.AttrTokenizer.AttrValueTokenizer1
    ValueTokenizer2 = TOKENIZER.AttrTokenizer.AttrValueTokenizer2
    ValueTokenizer3 = TOKENIZER.AttrTokenizer.AttrValueTokenizer3

    def test_highlight(self):
        doc = self._getdoc('<a b=c d="e" f=\'g\'>')
        doc.mode.run_tokenizer(None)
        
        kaa_testutils.check_style(doc, 0, 19, 
            [self.TOKENIZER.tokens.tag] * 2 +
            [self.AttrTokenizer.tokens.default] * 1 + 
            [self.AttrTokenizer.tokens.attr] * 2 +
            [self.ValueTokenizer3.tokens.value] * 1 +
            [self.AttrTokenizer.tokens.default] * 1 +
            [self.AttrTokenizer.tokens.attr] * 3 +
            [self.ValueTokenizer2.tokens.value] * 2 +
            [self.AttrTokenizer.tokens.default] * 1 +
            [self.AttrTokenizer.tokens.attr] * 3 +
            [self.ValueTokenizer1.tokens.value] * 2 +
            [self.TOKENIZER.tokens.tag] * 1)

    ValueJSTokenizer1 = TOKENIZER.AttrTokenizer.AttrValueJSTokenizer1
    ValueJSTokenizer2 = TOKENIZER.AttrTokenizer.AttrValueJSTokenizer2

    def test_jsattr(self):
        doc = self._getdoc("<a ona='if'>")
        doc.mode.run_tokenizer(None)

        kaa_testutils.check_style(doc, 0, 12, 
            [self.TOKENIZER.tokens.tag] * 2 +
            [self.AttrTokenizer.tokens.default] * 1 + 
            [self.AttrTokenizer.tokens.attr] * 5 +
            [self.ValueJSTokenizer1.tokens.keyword] * 2 +
            [self.AttrTokenizer.tokens.attr] * 1 +
            [self.TOKENIZER.tokens.tag] * 1)

        doc = self._getdoc('<a ona="if">')
        doc.mode.run_tokenizer(None)

        kaa_testutils.check_style(doc, 0, 12, 
            [self.TOKENIZER.tokens.tag] * 2 +
            [self.AttrTokenizer.tokens.default] * 1 + 
            [self.AttrTokenizer.tokens.attr] * 5 +
            [self.ValueJSTokenizer2.tokens.keyword] * 2 +
            [self.AttrTokenizer.tokens.attr] * 1 +
            [self.TOKENIZER.tokens.tag] * 1)

    JSTokenizer = TOKENIZER.JSTokenizer
    def test_jselem(self):
        doc = self._getdoc("<script>if</script>")

        doc.mode.run_tokenizer(None)
        kaa_testutils.check_style(doc, 0, 19, 
            [self.TOKENIZER.tokens.scripttag] * 8 +
            [self.JSTokenizer.tokens.keyword] * 2 +
            [self.TOKENIZER.tokens.closetag] * 9)


    AttrCSSTokenizer1 = TOKENIZER.AttrTokenizer.AttrValueCSSTokenizer1
    PropValueTokenizer1 = AttrCSSTokenizer1.PropValueTokenizer

    AttrCSSTokenizer2 = TOKENIZER.AttrTokenizer.AttrValueCSSTokenizer2
    PropValueTokenizer2 = AttrCSSTokenizer2.PropValueTokenizer

    def test_cssattr(self):
        doc = self._getdoc("<a style='a:b'>")
        doc.mode.run_tokenizer(None)

        kaa_testutils.check_style(doc, 0, 15, 
            [self.TOKENIZER.tokens.tag] * 2 +
            [self.AttrTokenizer.tokens.default] * 1 + 
            [self.AttrTokenizer.tokens.attr] * 7 +
            [self.AttrCSSTokenizer1.tokens.propname] * 2 +
            [self.PropValueTokenizer1.tokens.string] * 1 +
            [self.AttrTokenizer.tokens.attr] * 1 +
            [self.TOKENIZER.tokens.tag] * 1)


        doc = self._getdoc('<a style="a:b">')
        doc.mode.run_tokenizer(None)

        kaa_testutils.check_style(doc, 0, 15, 
            [self.TOKENIZER.tokens.tag] * 2 +
            [self.AttrTokenizer.tokens.default] * 1 + 
            [self.AttrTokenizer.tokens.attr] * 7 +
            [self.AttrCSSTokenizer2.tokens.propname] * 2 +
            [self.PropValueTokenizer2.tokens.string] * 1 +
            [self.AttrTokenizer.tokens.attr] * 1 +
            [self.TOKENIZER.tokens.tag] * 1)

        kaa_testutils.check_style(doc, 0, 15, 
            [self.TOKENIZER.tokens.tag] * 2 +
            [self.AttrTokenizer.tokens.default] * 1 + 
            [self.AttrTokenizer.tokens.attr] * 7 +
            [self.AttrCSSTokenizer2.tokens.propname] * 2 +
            [self.PropValueTokenizer2.tokens.string] * 1 +
            [self.AttrTokenizer.tokens.attr] * 1 +
            [self.TOKENIZER.tokens.tag] * 1)

    CSSTokenizer = TOKENIZER.CSSTokenizer
    CSSPropTokenizer = CSSTokenizer.PropTokenizer
    CSSPropValueTokenizer = CSSPropTokenizer.PropValueTokenizer
    def test_csselem(self):
        doc = self._getdoc("<style>a{b:c}</style>")
        doc.mode.run_tokenizer(None)
        kaa_testutils.check_style(doc, 0, 21, 
            [self.TOKENIZER.tokens.styletag] * 7 +
            [self.CSSTokenizer.tokens.default] * 1 +
            [self.CSSTokenizer.tokens.ruleset] * 1 +
            [self.CSSPropTokenizer.tokens.propname] * 2	 +
            [self.CSSPropValueTokenizer.tokens.string] * 1 +
            [self.CSSPropTokenizer.tokens.terminate_name] * 1 +
            [self.TOKENIZER.tokens.closetag] * 8 )


class TestEncDecl:

    def test_html5(self):
        assert 'UTF-8' == get_encoding(b'<meta charset="UTF-8">')
        assert 'UTF-8' == get_encoding(b"<meta charset='UTF-8'>")
        assert not get_encoding(b"<meta charset='UTF-8>")

    def test_html(self):
        assert 'UTF-8' == get_encoding(
            b'''<meta http-equiv="Content-type"
            content="text/html;charset=UTF-8">''')

        assert not get_encoding(
            b'''<meta con="text/html;charset=UTF-8">''')

        assert not get_encoding(
            b'''<meta content="text/html;char=UTF-8">''')

    def test_xml(self):
        assert 'UTF-8' == get_encoding(
            b'<?xml version="1.0" encoding="UTF-8"?>')

        assert not get_encoding(
            b'<?xm version="1.0" encoding="UTF-8"?>')

        assert not get_encoding(
            b'<?xml version="1.0" encoding=""?>')



from kaa.filetype.default import defaultmode
from kaa.keyboard import *
from kaa.syntax_highlight import *

CThemes = {
    'basic': []
}

c_keys = {
    (ctrl, 't'): 'toc.showlist',
}

KEYWORDS = ['asm', 'auto', 'break', 'case', 'continue', 'const', 'char',
            'default', 'do', 'double', 'else', 'endasm', 'entry', 'enum',
            'extern', 'float', 'for', 'goto', 'if', 'int', 'long', 'register',
            'return', 'short', 'signed', 'sizeof', 'static', 'struct',
            'switch', 'typedef', 'union', 'unsigned', 'void', 'volatile',
            'while']


CONSTANTS = ['NULL']


def make_tokenizer():
    return Tokenizer(tokens=[
        ("directive", SingleToken('directive', [r'\#\s*\w+'])),
        ("keyword", Keywords('keyword', KEYWORDS)),
        ("constant", Keywords('constant', CONSTANTS)),
        ("number", SingleToken('number',
                               [r'\b[0-9]+(\.[0-9]*)*\b', r'\b\.[0-9]+\b'])),
        ("comment1", Span('comment', r'/\*', '\*/', escape='\\')),
        ("comment2", Span('comment', r'//', '$', escape='\\')),

        ("string1", Span('string', '"', '"', escape='\\')),
        ("string2", Span('string', "'", "'", escape='\\')),
    ])


class CMode(defaultmode.DefaultMode):
    MODENAME = 'C'
    tokenizer = make_tokenizer()

    def init_keybind(self):
        super().init_keybind()
        self.register_keys(self.keybind, [c_keys])

    def init_themes(self):
        super().init_themes()
        self.themes.append(CThemes)



import kaa
from kaa.ui.msgbox import msgboxmode


class MenuMode(msgboxmode.MsgBoxMode):
    SEPARATOR = '/'
    USE_UNDO = False
    CLOSE_ON_DEL_WINDOW = True

    @classmethod
    def show_menu(cls, wnd, itemname, parentitems=None):

        mode = wnd.document.mode
        items = mode.get_menu(itemname)
        if not items:
            return

        items = [item for (item, submenu, commandid) in items
                 if not submenu or mode.get_menu(submenu)]

        doc = cls.build_msgbox('', items, lambda c: doc.mode._selected(c))
        doc.mode.target = wnd
        doc.mode.itemname = itemname
        doc.mode.parentitems = parentitems

        kaa.app.show_dialog(doc)
        return doc

    def close(self):
        super().close()
        self.target = None

    def _selected(self, c):
        if not c:
            if self.parentitems:
                MenuMode.show_menu(
                    self.target, self.parentitems[-1], self.parentitems[:-1])
            return

        itemname = self.shortcuts[c]
        menus = self.target.document.mode.get_menu(self.itemname)
        if not menus:
            return

        for item, submenu, commandid in menus:
            if itemname == item:
                if submenu:
                    parents = self.parentitems if self.parentitems else []
                    parents.append(self.itemname)

                    MenuMode.show_menu(self.target, submenu, parents)
                else:
                    self.target.document.mode.on_commands(
                        self.target, [commandid])



import os
import signal
import gc
import time
import sys
import select
import curses
import threading
import contextlib

import kaa
import kaa.log
from . import keydef, color, dialog
from kaa import clipboard
from kaa import document
from kaa import keyboard
from kaa import macro

from kaa.exceptions import KaaError


class CuiApp:
    DEFAULT_MENU_MESSAGE = 'Type F1 or alt+m for menu.'
    DEFAULT_THEME = 'basic'
    DEFAULT_PALETTE = 'dark'
    NUM_NEWFILE = 1

    def __init__(self, config):
        self.config = config
        self.clipboard = clipboard.select_clipboard()
        self.colors = None
        self._idleprocs = None
        self.lastcommands = ()
        self.focus = None
        self._quit = False
        self.theme = self.DEFAULT_THEME
        self.last_dir = '.'
        self._input_readers = []

        self._lock = threading.RLock()
        self._tasks = []

        self.commands = {}
        self.is_availables = {}

        self.sigwinch_rfd, self.sigwinch_wfd = os.pipe()
        signal.signal(signal.SIGWINCH,
                      lambda *args: os.write(self.sigwinch_wfd, b'0'))

    def register_commandobj(self, cmds):
        self.commands.update(cmds.get_commands())
        self.is_availables.update(cmds.get_commands_is_enable())

    def init_commands(self):
        from kaa.commands import appcommand, toolcommand, filecommand, gitcommand

        self.app_commands = appcommand.ApplicationCommands()
        self.register_commandobj(self.app_commands)

        self.file_commands = filecommand.FileCommands()
        self.register_commandobj(self.file_commands)

        self.register_commandobj(toolcommand.ToolCommands())
        self.register_commandobj(appcommand.MacroCommands())
        self.register_commandobj(gitcommand.GitCommands())

        for name in dir(self):
            attr = getattr(self, name)
            if hasattr(attr, 'COMMAND_ID') and callable(attr):
                self.commands[getattr(attr, 'COMMAND_ID')] = attr

    def init(self, mainframe):
        self.init_commands()

        if self.config.palette:
            self.set_palette(self.config.palette)
        elif not self.colors:
            self.set_palette(self.DEFAULT_PALETTE)

        self.config.init_history()

        from kaa.ui.messagebar import messagebarmode
        self.messagebar = messagebarmode.MessageBarMode()

        doc = document.Document()
        doc.setmode(self.messagebar)
        mainframe.set_messagebar(doc)

        self.mainframe = mainframe
        self.focus = self.mainframe
        self.macro = macro.Macro()

        self.mainframe.on_console_resized()
        self.messagebar.set_message(self.DEFAULT_MENU_MESSAGE)

    def on_shutdown(self):
        os.close(self.sigwinch_rfd)
        os.close(self.sigwinch_wfd)
        self.config.close()

    def get_current_theme(self):
        return self.theme

    def set_palette(self, name):
        palette = self.get_palette(name)
        self.colors = color.Colors(palette)

    def get_palette(self, name):
        if name == 'light':
            return color.LightPalette()
        else:
            return color.DarkPalette()

    def quit(self):
        self._quit = True

    def call_later(self, secs, f, *args, **kwargs):
        with self._lock:
            self._tasks.append((time.time() + secs, f, args, kwargs))

    SCHEDULE_WAIT_MARGIN = 0.05

    def _next_scheduled_task(self):
        with self._lock:
            tasks = sorted(t for t, f, a, k in self._tasks)
            if tasks:
                wait = max(0, tasks[0] - time.time())
                return wait + wait * self.SCHEDULE_WAIT_MARGIN

    def _run_scheduled_task(self):
        with self._lock:
            now = time.time()
            for n, (t, f, a, k) in enumerate(self._tasks):
                if t <= now:
                    del self._tasks[n]
                    f(*a, **k)
                    return

    def get_command(self, commandid):
        cmd = self.commands.get(commandid, None)
        if cmd:
            is_available = self.is_availables.get(commandid, None)
            return (is_available, cmd)

    def set_idlejob(self):
        self._idleprocs = [
            doc.mode.on_idle for doc in document.Document.all if doc.mode]

    def on_idle(self):
        if self._idleprocs:
            proc = self._idleprocs.pop(0)
            # proc() returns True if proc() still has job to be done.
            if proc():
                self._idleprocs.append(proc)

            return True
        else:
            return False

    def translate_theme(self, theme):
        overlays = {}
        for name, overlay in theme.overlays.items():
            fg = bg = None
            if overlay.fgcolor:
                fg = self.colors.colornames.get(overlay.fgcolor.upper())
            if overlay.bgcolor:
                bg = self.colors.colornames.get(overlay.bgcolor.upper())
            overlays[name] = (fg, bg)

        for style in theme.styles.values():
            fg, bg = (self.colors.colornames.get(style.fgcolor.upper()),
                      self.colors.colornames.get(style.bgcolor.upper()))
            attr = self.colors.get_color(fg, bg)
            style.cui_colorattr = attr

            style.cui_overlays = {}
            for name, (o_fg, o_bg) in overlays.items():
                if o_fg is None:
                    o_fg = fg
                if o_bg is None:
                    o_bg = bg
                style.cui_overlays[name] = self.colors.get_color(o_fg, o_bg)

    def get_keyname(self, key):
        try:
            if not isinstance(key, int):
                key = ord(key)

            return str(curses.keyname(key), 'utf-8', 'replace')
        except Exception:
            return '?'

    def translate_key(self, mod, c):
        """Translate kaa's key value to curses keycode"""

        alt = keyboard.alt in mod
        ctrl = keyboard.ctrl in mod
        shift = keyboard.shift in mod

        if alt:
            meta = '\x1b'
        else:
            meta = ''

        if isinstance(c, str):
            if shift:
                raise KaaError(
                    'Cannot use shift key for character: {!r}'.format((mod, c)))
            if ctrl:
                c = c.upper()
                if not (0x40 <= ord(c) <= 0x5f):
                    raise KaaError(
                        'Cannot use control key for character: {!r}'.format((mod, c)))
                return meta + chr(ord(c) - 0x40)
            else:
                return meta + c
        else:
            ret = keydef.keyfromname(c, ctrl, shift)
            if ret is None:
                raise KaaError(
                    'Cannot convert character: {!r}'.format((mod, c)))

            return [ret] if not meta else [meta, ret]

    def set_focus(self, wnd):
        if wnd is self.focus:
            return

        if self.focus:
            self.focus.on_killfocus()

        self.focus = wnd
        if wnd:
            wnd.on_focus()

    def show_doc(self, doc):
        '''
        Create new window for the doc and show it.
        '''
        ret = self.mainframe.show_doc(doc)
        return ret

    def show_inputline(self, doc):
        self._idleprocs = None  # Reschedule idle procs
        dlg = dialog.DialogWnd(parent=self.mainframe, doc=doc)
        self.mainframe.show_inputline(dlg)
        return dlg

    def show_dialog(self, doc):
        dlg = dialog.DialogWnd(parent=self.mainframe, doc=doc)
        self.mainframe.show_dialog(dlg)
        return dlg

    def get_frames(self):
        return self.mainframe.childframes[:]

    def get_activeframe(self):
        return self.mainframe.activeframe

    def show_cursor(self, f):
        try:
            curses.curs_set(f)
        except curses.error:
            # curses.curs_set() occasionally fails if $TERM=xterm-color
            pass

    @contextlib.contextmanager
    def restore_teminal(self):
        curses.def_prog_mode()
        curses.endwin()
        try:
            yield
        finally:
            curses.reset_prog_mode()
            self.mainframe.refresh()

    def add_input_reader(self, reader):
        self._input_readers.append(reader)

    def del_input_reader(self, reader):
        if reader in self._input_readers:
            self._input_readers.remove(reader)

    def dump_panel(self):
        import curses.panel
        panels = []
        p = curses.panel.top_panel()
        while p:
            panels.append(p)
            p = p.below()

        for w in self.mainframe.walk_children():
            idx = panels.index(w._panel)
            d = getattr(w, 'document', w)
            m = getattr(d, 'mode', w)
            panels[idx] = m

    def run(self):
        #        def f(t, i):
        #            _trace(t, i)
        #        gc.callbacks.append(f)
        gc.set_threshold(2000, 10, 10)

        nonblocking = True
        while not self._quit:
            try:
                if not self.focus:
                    kaa.log.error('Internal error: invalid focus window.')
                    break

                self.focus.restore_cursor_pos()
                if not nonblocking:
                    # update screen before sleep.
                    curses.panel.update_panels()
                    curses.doupdate()

                rd = []
                for f in self._input_readers:
                    rd.extend(f.get_reader())

                try:
                    rlist, _, _ = select.select(
                        [sys.stdin, self.sigwinch_rfd] + rd, [], [],
                        0 if nonblocking else self._next_scheduled_task())

                except InterruptedError:
                    pass

                if not nonblocking and not rlist:
                    # timeout
                    self._run_scheduled_task()
                    self.set_idlejob()  # Reschedule idle procs

                if self.sigwinch_rfd in rlist:
                    os.read(self.sigwinch_rfd, 1)

                    # sigh. pep-0475 prevents us to handling SIGWINCH.
                    # force curses to resize window.
                    import struct
                    import fcntl
                    import termios
                    v = fcntl.ioctl(0, termios.TIOCGWINSZ,
                                    struct.pack('HHHH', 0, 0, 0, 0))
                    lines, cols, _, _ = struct.unpack('HHHH', v)
                    curses.resizeterm(lines, cols)
                    self.mainframe.on_console_resized()

                ready = [r for r in rlist if r not in
                         (sys.stdin, self.sigwinch_rfd)]
                if ready:
                    nonblocking = True
                    for r in ready:
                        idx = rd.index(r)
                        self._input_readers[idx].read_input(r)
                    self.set_idlejob()  # Reschedule idle procs

                inputs = self.focus.do_input(nonblocking=True)
                for c in inputs:
                    if isinstance(c, keydef.KeyEvent):
                        nonblocking = True
                        if c.key == curses.KEY_RESIZE:
                            self.mainframe.on_console_resized()
                            continue

                        if self.focus.editmode:
                            self.focus.editmode.on_keyevent(self.focus, c)
                            if self.focus:
                                self.focus.update_window()

                if not inputs:
                    if self.mainframe.on_idle():
                        self.set_idlejob()  # Reschedule idle procs
                        continue

                    # no input
                    if not self.on_idle():
                        # No more idle jobs
                        nonblocking = False
                else:
                    self.set_idlejob()  # Reschedule idle procs

            except Exception as e:
                kaa.log.error('Unhandled exception', exc_info=True)
                kaa.app.messagebar.set_message(' '.join(str(e).split('\n')))

                nonblocking = True



import kaa
from kaa.ui.dialog import dialogmode
from kaa.ui.selectlist import filterlist, selectlist
from kaa.command import commandid, norec, norerun
from kaa.keyboard import *


class TOCList(filterlist.FilterListMode):
    MAX_CAPTION_LEN = 30
    USE_PHRASE_STYLE = True

    def set_candidates(self, candidates):
        self.candidates = [
            selectlist.SelectItem(
                'selectitem' if c.token != 'namespace' else 'selectitem2', 'selectitem-active', c.dispname, c)
            for c in candidates]

    def _filter_items(self, query):
        first = None
        if query:
            query = query.upper().split()
            matches = self.candidates[:]
            for i, s in enumerate(self.candidates):
                u = s.value.name.upper()
                for q in query:
                    if q not in u:
                        matches[i] = False
                        break
                    if not first:
                        first = s

            ns = None
            for i in range(len(self.candidates)):
                if self.candidates[i].value.token == 'namespace':
                    ns = i
                else:
                    if matches[i] and ns is not None:
                        matches[ns] = self.candidates[ns]
            items = [m for m in matches if m]
        else:
            items = self.candidates[:]
            first = items[0] if items else None

        return first, items

    def update_doc(self, items):
        self.items = list(items)
        self.cursel = None

        self.document.marks.clear()
        self.document.delete(0, self.document.endpos())
        with dialogmode.FormBuilder(self.document) as f:
            prev = None
            for item in self.items:
                if item.value.token == 'namespace':
                    if prev:
                        f.append_text('default',
                                      '\n' + '  ' * len(item.value.get_parents()))
                    f.append_text(item.style, item.text, mark_pair=item)

                else:
                    if prev and (prev.value.token == 'namespace' or
                                 prev.value.get_parents() != item.value.get_parents()):

                        f.append_text('default',
                                      '\n' + '  ' * (len(item.value.get_parents())))

                    f.append_text(item.style, item.text, mark_pair=item)
                    f.append_text('default', ' ')

                prev = item

    def sel_next_namespace(self, wnd):
        if not self.items:
            newsel = None
        elif self.cursel is None:
            newsel = self.items[0]
        else:
            try:
                idx = self.items.index(self.cursel)
            except ValueError:
                newsel = self.items[0]
            else:
                idx += 1
                while idx < len(self.items):
                    newsel = self.items[idx]
                    if newsel.value.token == 'namespace':
                        break
                    idx += 1
                else:
                    newsel = self.items[0]

        self.update_sel(wnd, newsel, bottom=True)
        return newsel

    def sel_prev_namespace(self, wnd):
        bottom = None
        if not self.items:
            newsel = None
        elif self.cursel is None:
            newsel = self.items[-1]
            bottom = True
        else:
            newsel = None
            for i, item in enumerate(self.items):
                if item is self.cursel:
                    bottom = True
                    if newsel:
                        break
                    newsel = None

                if item.value.token == 'namespace':
                    newsel = item

            if not newsel:
                newsel = self.items[-1]

        self.update_sel(wnd, newsel, bottom=bottom)
        return newsel


toclistdlg_keys = {
    tab: 'filterlistdlg.next',
    (shift, tab): 'filterlistdlg.prev',
    (alt, 'p'): 'toclistdlg.prev_namespace',
    (alt, 'n'): 'toclistdlg.next_namespace',
}


class TOCListInputDlgMode(filterlist.FilterListInputDlgMode):

    def init_keybind(self):
        super().init_keybind()
        self.keybind.add_keybind(toclistdlg_keys)

    @commandid('toclistdlg.next_namespace')
    @norec
    @norerun
    def sel_next_namespace(self, wnd):
        filterlist = wnd.get_label('filterlist')
        filterlist.document.mode.sel_next_namespace(filterlist)

    @commandid('toclistdlg.prev_namespace')
    @norec
    @norerun
    def sel_prev_namespace(self, wnd):
        filterlist = wnd.get_label('filterlist')
        filterlist.document.mode.sel_prev_namespace(filterlist)


def show_toclist(wnd, toclist):
    def callback(result):
        if result:
            wnd.cursor.setpos(result.pos)

    doc = TOCListInputDlgMode.build(
        'Search:', callback)
    dlg = kaa.app.show_dialog(doc)

    filterlistdoc = TOCList.build()
    list = dlg.add_doc('dlg_filterlist', 0, filterlistdoc)

    filterlistdoc.mode.set_candidates(toclist)
    filterlistdoc.mode.set_query(list, '')
    dlg.on_console_resized()

    pos = wnd.cursor.pos
    last = None
    for item in filterlistdoc.mode.candidates:
        if pos < item.value.pos:
            break
        last = item

    if last:
        filterlistdoc.mode.update_sel(list, last, bottom=True)

    return dlg



import functools
import kaa
from kaa.command import Commands
from kaa.command import commandid
from kaa.command import norec
from kaa.command import norerun


class ViCommands(Commands):

    @commandid('edit.replace-next-char')
    @norec
    @norerun
    def replace_next_char(self, wnd):
        wnd.editmode.install_keyhook(self._hook_replacechar)

    def _hook_replacechar(self, wnd, keyevent):
        s = keyevent.key[0]
        if isinstance(s, str) and s >= ' ':
            wnd.document.mode.put_string(wnd, s, overwrite=True)
            return None
        return keyevent

    @commandid('edit.delete-next-move')
    @norec
    @norerun
    def delete_next_move(self, wnd):
        wnd.editmode.install_keyhook(self._hook_key_delete_next_move)

    def _hook_key_delete_next_move(self, wnd, keyevent):
        s = keyevent.key
        if isinstance(s, str) and s == 'd':
            wnd.document.mode.on_commands(wnd, ['edit.delete.currentline'])
        else:
            f = functools.partial(self._hook_delete_next_move,
                                  wnd.cursor.pos)
            wnd.editmode.install_post_key_hook(f)
            return keyevent

    def _hook_delete_next_move(self, posfrom, wnd, s, commands):
        if kaa.app.focus is wnd:
            if commands:
                f, t = sorted((posfrom, wnd.cursor.pos))
                if f != t and t <= wnd.document.endpos():
                    wnd.document.mode.delete_string(wnd, f, t)





rec_command = object()
rec_string = object()


class Macro:
    recording = False
    commands = ()

    def start_record(self):
        self.recording = True
        self.commands = []

    def end_record(self):
        self.recording = False

    def toggle_record(self):
        self.recording = not self.recording
        if self.recording:
            self.commands = []

    def is_recording(self):
        return self.recording

    def get_commands(self):
        return self.commands

    def record(self, n_repeat, commandid, *args, **kwargs):
        if not isinstance(commandid, str):
            if hasattr(commandid, 'NOREC'):
                return

            commandid = commandid.COMMAND_ID

        self.commands.append((rec_command, n_repeat, commandid, args, kwargs))

    def record_string(self, s, overwrite):
        self.commands.append([rec_string, overwrite, s])

    def run(self, wnd):
        if self.recording:
            return

        mode = wnd.document.mode
        for rec in self.commands:
            if rec[0] is rec_string:
                wnd.document.mode.put_string(wnd, rec[2], overwrite=rec[1])
                wnd.screen.selection.clear()
            else:
                n_repeat, commandid, args, kwargs = rec[1:]
                wnd.set_command_repeat(n_repeat)
                try:
                    if callable(commandid):
                        commandid(wnd)
                    else:
                        is_available, cmd = mode.get_command(commandid)
                        cmd(wnd, *args, **kwargs)
                finally:
                    wnd.set_command_repeat(1)



import os
from kaa.command import Commands
from kaa.command import commandid
from kaa.command import norec
from kaa.command import norerun


class GitCommands(Commands):

    def _get_cur_dir(self, wnd):
        if wnd.document.fileinfo:
            return wnd.document.fileinfo.dirname
        else:
            return os.getcwd()

    @commandid('git.status')
    @norec
    @norerun
    def status(self, wnd):
        from kaa.ui.git import statusmode
        statusmode.show_git_status(self._get_cur_dir(wnd))

    @commandid('git.log')
    @norec
    @norerun
    def log(self, wnd):
        from kaa.ui.git import logmode
        logmode.show_git_log(self._get_cur_dir(wnd))
